import	os
import	unittest
from	pygwarts.hedwig.mail.letter.fields	import SenderField
from	pygwarts.hedwig.mail.letter.fields	import RecipientField
from	pygwarts.hedwig.mail.letter.fields	import ccField
from	pygwarts.hedwig.mail.letter.fields	import bccField
from	pygwarts.hedwig.mail.letter.fields	import SubjectField
from	pygwarts.hedwig.mail.letter.fields	import BodyField
from	pygwarts.hedwig.mail.letter.fields	import AttachmentField
from	pygwarts.hedwig.mail.builder.client	import ThunderbirdBuilder
from	pygwarts.hedwig.mail.builder.client	import OutlookBuilder
from	pygwarts.hedwig.mail.utils			import EmailValidator
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.irma.access.utils			import TextWrapper
from	pygwarts.tests.hedwig				import HedwigTestCase








class ClientCases(HedwigTestCase):

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.HEDWIG_MAIL_CLIENT_HANDLER): os.remove(cls.HEDWIG_MAIL_CLIENT_HANDLER)

		# IMPORTANT!
		# IF CLIENT TESTS TO BE MANUALLY
		# INVOKED, BELOW REMOVES WILL NOT
		# ALLOW ATTACHMENTS TO BE INCLUDED
		if	os.path.isfile(cls.attachyone):	os.remove(cls.attachyone)
		if	os.path.isfile(cls.attachytwo):	os.remove(cls.attachytwo)

	@classmethod
	def setUpClass(cls):

		cls.make_loggy_file(cls, cls.HEDWIG_MAIL_CLIENT_HANDLER)
		cls.attachyone = str(cls.HEDWIG_ROOT /"atachy.one")
		cls.attachytwo = str(cls.HEDWIG_ROOT /"atachy.two")
		if	not os.path.isfile(cls.attachyone): cls.fmake(cls, cls.attachyone, "heil maximum build!")
		if	not os.path.isfile(cls.attachytwo): cls.fmake(cls, cls.attachytwo, "heil maximum build!")

	@unittest.skip("manual invokation only")
	def test_Thunderbird_build_minimal(self):

		class CurrentBuilder(ThunderbirdBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_CLIENT_HANDLER
				init_name	= "ThunderbirdBuilder_minimal"
				init_level	= 10

			class validator(EmailValidator):	pass
			class one(SenderField):				field_value = "example1@email.com"
			class two(RecipientField):			field_value = "example2@email.com", "example3@email.com"
			@TextWrapper("Subject: ")
			class five(SubjectField):

				field_value	= "{level} build test"
				modifiers	= { "level": "maximum" }

			@TextWrapper(header="My ", footer=", baby!")
			class six(BodyField):

				field_value	= "congratulations! Build is {quality}"
				modifiers	= { "quality": "nice" }

		self.test_case = CurrentBuilder()
		current = self.test_case.build()
		self.assertIsInstance(current, dict)
		self.assertEqual(len(current), 7)
		self.assertIn("sender", current)
		self.assertIn("recipient", current)
		self.assertIn("cc", current)
		self.assertIn("bcc", current)
		self.assertIn("subject", current)
		self.assertIn("body", current)
		self.assertIn("attachment", current)




	@unittest.skip("manual invokation only")
	def test_Thunderbird_build_full(self):

		class CurrentBuilder(ThunderbirdBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_CLIENT_HANDLER
				init_name	= "ThunderbirdBuilder_full"
				init_level	= 10

			class validator(EmailValidator):	pass
			class s1(SenderField):				field_value = "example1@email.com"
			class s2(SenderField):				field_value = "exomple1@email.com"
			class two(RecipientField):			field_value = "example2@email.com", "example3@email.com"
			class three(ccField):				field_value = "example4@email.com", "example5@email.com"
			class four(bccField):				field_value = "example6@email.com", "example7@email.com"
			@TextWrapper("Subject: ")
			class five(SubjectField):

				field_value	= "{level} build test"
				modifiers	= { "level": "maximum" }

			@TextWrapper(header="My ", footer=", baby!")
			class six(BodyField):

				field_value	= "congratulations! Build is {quality}"
				modifiers	= { "quality": "nice" }

			class a1(AttachmentField): field_value = self.attachyone
			class a2(AttachmentField):

				field_value	= "attachytwo"
				modifiers	= lambda *S : self.attachytwo

		self.test_case = CurrentBuilder()
		current = self.test_case.build()
		self.assertIsInstance(current, dict)
		self.assertEqual(len(current), 7)
		self.assertIn("sender", current)
		self.assertIn("recipient", current)
		self.assertIn("cc", current)
		self.assertIn("bcc", current)
		self.assertIn("subject", current)
		self.assertIn("body", current)
		self.assertIn("attachment", current)








	@unittest.skip("manual invokation only")
	def test_Outlook_build_nomodule(self):

		class CurrentBuilder(OutlookBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_CLIENT_HANDLER
				init_name	= "OutlookBuilder_nomodule"
				init_level	= 10

			class validator(EmailValidator):	pass
			class one(SenderField):				field_value = "example1@email.com"
			class two(RecipientField):			field_value = "example2@email.com", "example3@email.com"

		with self.assertLogs("OutlookBuilder_nomodule", 20) as case_loggy:

			self.test_case = CurrentBuilder()
			current = self.test_case.build()

		self.assertEqual(len(case_loggy.output),1)
		self.assertIn(f"INFO:OutlookBuilder_nomodule:{self.test_case} is not applicable", case_loggy.output)




	@unittest.skip("manual invokation only")
	def test_Outlook_build_minimal(self):

		class CurrentBuilder(OutlookBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_CLIENT_HANDLER
				init_name	= "OutlookBuilder_minimal"
				init_level	= 10

			class validator(EmailValidator):	pass
			class s1(SenderField):				field_value = "example1@email.com"
			class two(RecipientField):			field_value = "example2@email.com", "example3@email.com"
			@TextWrapper("Subject: ")
			class five(SubjectField):

				field_value	= "{level} build test"
				modifiers	= { "level": "maximum" }

			@TextWrapper(header="My ", footer=", baby!")
			class six(BodyField):

				field_value	= "congratulations! Build is {quality}"
				modifiers	= { "quality": "nice" }

		self.test_case = CurrentBuilder()
		current = self.test_case.build()
		self.assertIsInstance(current, dict)
		self.assertEqual(len(current), 7)
		self.assertIn("sender", current)
		self.assertIn("recipient", current)
		self.assertIn("cc", current)
		self.assertIn("bcc", current)
		self.assertIn("subject", current)
		self.assertIn("body", current)
		self.assertIn("attachment", current)




	@unittest.skip("manual invokation only")
	def test_Outlook_build_full(self):

		class CurrentBuilder(OutlookBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_CLIENT_HANDLER
				init_name	= "OutlookBuilder_full"
				init_level	= 10

			class validator(EmailValidator):	pass
			class s1(SenderField):				field_value = "example1@email.com"
			class s2(SenderField):				field_value = "exomple1@email.com"
			class two(RecipientField):			field_value = "example2@email.com", "example3@email.com"
			class three(ccField):				field_value = "example4@email.com", "example5@email.com"
			class four(bccField):				field_value = "example6@email.com", "example7@email.com"
			@TextWrapper("Subject: ")
			class five(SubjectField):

				field_value	= "{level} build test"
				modifiers	= { "level": "maximum" }

			@TextWrapper(header="My ", footer=", baby!")
			class six(BodyField):

				field_value	= "congratulations! Build is {quality}"
				modifiers	= { "quality": "nice" }

			class a1(AttachmentField): field_value = self.attachyone
			class a2(AttachmentField):

				field_value	= "attachytwo"
				modifiers	= lambda *S : self.attachytwo

		self.test_case = CurrentBuilder()
		current = self.test_case.build()
		self.assertIsInstance(current, dict)
		self.assertEqual(len(current), 7)
		self.assertIn("sender", current)
		self.assertIn("recipient", current)
		self.assertIn("cc", current)
		self.assertIn("bcc", current)
		self.assertIn("subject", current)
		self.assertIn("body", current)
		self.assertIn("attachment", current)








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	os
import	unittest
from	pathlib								import Path
from	shutil								import copyfile
from	shutil								import rmtree
from	time								import sleep
from	pygwarts.magical.philosophers_stone	import Transmutable
from	pygwarts.tests.hagrid				import EasySet
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.hagrid.thrivables			import Tree
from	pygwarts.hagrid.sprouts				import fssprout
from	pygwarts.hagrid.planting			import Flourish
from	pygwarts.hagrid.planting.peels		import GrowingPeel
from	pygwarts.hagrid.planting.leafs		import LeafPush
from	pygwarts.hagrid.bloom.leafs			import Rejuvenation








class LeafPushCases(EasySet):

	"""
		GrowingPeel, Rejuvenation, pushing grow
		Single sprout
		Single bough
	"""

	@classmethod
	def tearDownClass(cls):
		cls.clean(cls)

		if	cls.clean_up:
			if	os.path.isfile(cls.EASY_HANDLER_12): os.remove(cls.EASY_HANDLER_12)

	@staticmethod
	def crossfs_copy(*args, **kwargs):

		copyfile(*args, **kwargs)
		raise	PermissionError("[Errno 1] Operation not permitted")

	@staticmethod
	def dontlike_copy(*args, **kwargs):
		raise	ValueError(f"I don't like \"{args[0]}\"")

	@staticmethod
	def empty_copy(*args, **kwargs):	pass

	@classmethod
	def setUpClass(cls):

		super().setUpClass()
		cls.make_loggy_file(cls, cls.EASY_HANDLER_12)

	def setUp(self):
		class Sakura(Tree):

			bough = self.EASY_SET_BOUGH
			class loggy(LibraryContrib):

				handler		= self.EASY_HANDLER_12
				init_level	= 10

			@GrowingPeel
			class grow(LeafPush):		pass
			class files(Rejuvenation):	pass

			@fssprout(self.EASY_SET_SPROUT)
			class sync(Flourish):		pass

		self.fmake(self.file1, "OOH EEH OOH AH AH TING TANG WALA WALA BING BANG")
		self.fmake(self.pros_file1, "may be ain't best way")
		self.fmake(self.pros_file2, "probably the best way")
		self.fmake(self.pros_file3, "definitely the best way")
		self.fmake(self.cons_file1, "might cause a headache")
		self.fmake(self.cons_file2, "annihilation")
		self.fmake(self.redundant_1, "no use 1")
		self.fmake(self.redundant_2, "no use 2")
		self.test_case = Sakura




	def test_push_a_first_flourish(self):

		sleep(1.5)
		self.test_case.loggy.init_name = "push_first_flourish"


		self.assertTrue(os.path.isfile(self.file1))
		self.assertTrue(os.path.isdir(self.pros_folder))
		self.assertTrue(os.path.isfile(self.pros_file1))
		self.assertTrue(os.path.isfile(self.pros_file2))
		self.assertTrue(os.path.isfile(self.pros_file3))
		self.assertTrue(os.path.isdir(self.cons_folder))
		self.assertTrue(os.path.isfile(self.cons_file1))
		self.assertTrue(os.path.isfile(self.cons_file2))
		self.assertTrue(os.path.isdir(self.redundant_1_folder))
		self.assertTrue(os.path.isdir(self.redundant_2_folder))
		self.assertTrue(os.path.isfile(self.redundant_1))
		self.assertTrue(os.path.isfile(self.redundant_2))


		file1_mt = os.path.getmtime(self.file1)
		pros_folder_mt = os.path.getmtime(self.pros_folder)
		pros_file1_mt = os.path.getmtime(self.pros_file1)
		pros_file2_mt = os.path.getmtime(self.pros_file2)
		pros_file3_mt = os.path.getmtime(self.pros_file3)
		cons_folder_mt = os.path.getmtime(self.cons_folder)
		cons_file1_mt = os.path.getmtime(self.cons_file1)
		cons_file2_mt = os.path.getmtime(self.cons_file2)
		redundant_1_folder_mt = os.path.getmtime(self.redundant_1_folder)
		redundant_2_folder_mt = os.path.getmtime(self.redundant_2_folder)
		redundant_1_mt = os.path.getmtime(self.redundant_1)
		redundant_2_mt = os.path.getmtime(self.redundant_2)


		self.assertFalse(os.path.isfile(self.dst_file1))
		self.assertFalse(os.path.isdir(self.dst_pros_folder))
		self.assertFalse(os.path.isfile(self.dst_pros_file1))
		self.assertFalse(os.path.isfile(self.dst_pros_file2))
		self.assertFalse(os.path.isfile(self.dst_pros_file3))
		self.assertFalse(os.path.isdir(self.dst_cons_folder))
		self.assertFalse(os.path.isfile(self.dst_cons_file1))
		self.assertFalse(os.path.isfile(self.dst_cons_file2))
		self.assertFalse(os.path.isdir(self.dst_redundant_1_folder))
		self.assertFalse(os.path.isdir(self.dst_redundant_2_folder))
		self.assertFalse(os.path.isfile(self.dst_redundant_1))
		self.assertFalse(os.path.isfile(self.dst_redundant_2))


		with self.assertLogs("push_first_flourish", 10) as case_loggy:

			self.test_case = self.test_case()
			self.test_case.sync()


		self.no_loggy_levels(case_loggy.output, 30,40,50)


		self.assertFalse(os.path.isfile(self.file1))
		self.assertTrue(os.path.isdir(self.pros_folder))
		self.assertFalse(os.path.isfile(self.pros_file1))
		self.assertFalse(os.path.isfile(self.pros_file2))
		self.assertFalse(os.path.isfile(self.pros_file3))
		self.assertTrue(os.path.isdir(self.cons_folder))
		self.assertFalse(os.path.isfile(self.cons_file1))
		self.assertFalse(os.path.isfile(self.cons_file2))
		self.assertTrue(os.path.isdir(self.redundant_1_folder))
		self.assertTrue(os.path.isdir(self.redundant_2_folder))
		self.assertFalse(os.path.isfile(self.redundant_1))
		self.assertFalse(os.path.isfile(self.redundant_2))


		self.assertTrue(os.path.isfile(self.dst_file1))
		self.assertTrue(os.path.isdir(self.dst_pros_folder))
		self.assertTrue(os.path.isfile(self.dst_pros_file1))
		self.assertTrue(os.path.isfile(self.dst_pros_file2))
		self.assertTrue(os.path.isfile(self.dst_pros_file3))
		self.assertTrue(os.path.isdir(self.dst_cons_folder))
		self.assertTrue(os.path.isfile(self.dst_cons_file1))
		self.assertTrue(os.path.isfile(self.dst_cons_file2))
		self.assertTrue(os.path.isdir(self.dst_redundant_1_folder))
		self.assertTrue(os.path.isdir(self.dst_redundant_2_folder))
		self.assertTrue(os.path.isfile(self.dst_redundant_1))
		self.assertTrue(os.path.isfile(self.dst_redundant_2))


		self.assertIn(

			f"INFO:push_first_flourish:Pushed leaf \"{self.dst_file1}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_first_flourish:Pushed leaf \"{self.dst_pros_file1}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_first_flourish:Pushed leaf \"{self.dst_pros_file2}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_first_flourish:Pushed leaf \"{self.dst_pros_file3}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_first_flourish:Pushed leaf \"{self.dst_cons_file1}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_first_flourish:Pushed leaf \"{self.dst_cons_file2}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_first_flourish:Pushed leaf \"{self.dst_redundant_1}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_first_flourish:Pushed leaf \"{self.dst_redundant_2}\"",
			case_loggy.output
		)


		self.assertEqual(int(file1_mt),int(os.path.getmtime(self.dst_file1)))
		self.assertEqual(int(pros_file1_mt),int(os.path.getmtime(self.dst_pros_file1)))
		self.assertEqual(int(pros_file2_mt),int(os.path.getmtime(self.dst_pros_file2)))
		self.assertEqual(int(pros_file3_mt),int(os.path.getmtime(self.dst_pros_file3)))
		self.assertEqual(int(cons_file1_mt),int(os.path.getmtime(self.dst_cons_file1)))
		self.assertEqual(int(cons_file2_mt),int(os.path.getmtime(self.dst_cons_file2)))
		self.assertEqual(int(redundant_1_mt),int(os.path.getmtime(self.dst_redundant_1)))
		self.assertEqual(int(redundant_2_mt),int(os.path.getmtime(self.dst_redundant_2)))








	def test_push_b_interval_flourish(self):

		sleep(1.5)
		self.test_case.loggy.init_name = "push_interval_flourish"


		file1_mt = os.path.getmtime(self.file1)
		pros_folder_mt = os.path.getmtime(self.pros_folder)
		pros_file1_mt = os.path.getmtime(self.pros_file1)
		pros_file2_mt = os.path.getmtime(self.pros_file2)
		pros_file3_mt = os.path.getmtime(self.pros_file3)
		cons_folder_mt = os.path.getmtime(self.cons_folder)
		cons_file1_mt = os.path.getmtime(self.cons_file1)
		cons_file2_mt = os.path.getmtime(self.cons_file2)
		redundant_1_folder_mt = os.path.getmtime(self.redundant_1_folder)
		redundant_2_folder_mt = os.path.getmtime(self.redundant_2_folder)
		redundant_1_mt = os.path.getmtime(self.redundant_1)
		redundant_2_mt = os.path.getmtime(self.redundant_2)


		self.assertTrue(os.path.isfile(self.file1))
		self.assertTrue(os.path.isdir(self.pros_folder))
		self.assertTrue(os.path.isfile(self.pros_file1))
		self.assertTrue(os.path.isfile(self.pros_file2))
		self.assertTrue(os.path.isfile(self.pros_file3))
		self.assertTrue(os.path.isdir(self.cons_folder))
		self.assertTrue(os.path.isfile(self.cons_file1))
		self.assertTrue(os.path.isfile(self.cons_file2))
		self.assertTrue(os.path.isdir(self.redundant_1_folder))
		self.assertTrue(os.path.isdir(self.redundant_2_folder))
		self.assertTrue(os.path.isfile(self.redundant_1))
		self.assertTrue(os.path.isfile(self.redundant_2))


		self.assertTrue(os.path.isfile(self.dst_file1))
		self.assertTrue(os.path.isdir(self.dst_pros_folder))
		self.assertTrue(os.path.isfile(self.dst_pros_file1))
		self.assertTrue(os.path.isfile(self.dst_pros_file2))
		self.assertTrue(os.path.isfile(self.dst_pros_file3))
		self.assertTrue(os.path.isdir(self.dst_cons_folder))
		self.assertTrue(os.path.isfile(self.dst_cons_file1))
		self.assertTrue(os.path.isfile(self.dst_cons_file2))
		self.assertTrue(os.path.isdir(self.dst_redundant_1_folder))
		self.assertTrue(os.path.isdir(self.dst_redundant_2_folder))
		self.assertTrue(os.path.isfile(self.dst_redundant_1))
		self.assertTrue(os.path.isfile(self.dst_redundant_2))


		with self.assertLogs("push_interval_flourish", 10) as case_loggy:

			self.test_case = self.test_case()
			self.test_case.sync()


		self.no_loggy_levels(case_loggy.output, 30,40,50)


		self.assertFalse(os.path.isfile(self.file1))
		self.assertTrue(os.path.isdir(self.pros_folder))
		self.assertFalse(os.path.isfile(self.pros_file1))
		self.assertFalse(os.path.isfile(self.pros_file2))
		self.assertFalse(os.path.isfile(self.pros_file3))
		self.assertTrue(os.path.isdir(self.cons_folder))
		self.assertFalse(os.path.isfile(self.cons_file1))
		self.assertFalse(os.path.isfile(self.cons_file2))
		self.assertTrue(os.path.isdir(self.redundant_1_folder))
		self.assertTrue(os.path.isdir(self.redundant_2_folder))
		self.assertFalse(os.path.isfile(self.redundant_1))
		self.assertFalse(os.path.isfile(self.redundant_2))


		self.assertTrue(os.path.isfile(self.dst_file1))
		self.assertTrue(os.path.isdir(self.dst_pros_folder))
		self.assertTrue(os.path.isfile(self.dst_pros_file1))
		self.assertTrue(os.path.isfile(self.dst_pros_file2))
		self.assertTrue(os.path.isfile(self.dst_pros_file3))
		self.assertTrue(os.path.isdir(self.dst_cons_folder))
		self.assertTrue(os.path.isfile(self.dst_cons_file1))
		self.assertTrue(os.path.isfile(self.dst_cons_file2))
		self.assertTrue(os.path.isdir(self.dst_redundant_1_folder))
		self.assertTrue(os.path.isdir(self.dst_redundant_2_folder))
		self.assertTrue(os.path.isfile(self.dst_redundant_1))
		self.assertTrue(os.path.isfile(self.dst_redundant_2))


		self.assertIn(

			f"INFO:push_interval_flourish:Pushed leaf \"{self.dst_file1}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_interval_flourish:Pushed leaf \"{self.dst_pros_file1}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_interval_flourish:Pushed leaf \"{self.dst_pros_file2}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_interval_flourish:Pushed leaf \"{self.dst_pros_file3}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_interval_flourish:Pushed leaf \"{self.dst_cons_file1}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_interval_flourish:Pushed leaf \"{self.dst_cons_file2}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_interval_flourish:Pushed leaf \"{self.dst_redundant_1}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_interval_flourish:Pushed leaf \"{self.dst_redundant_2}\"",
			case_loggy.output
		)


		self.assertEqual(int(file1_mt),int(os.path.getmtime(self.dst_file1)))
		self.assertEqual(int(pros_file1_mt),int(os.path.getmtime(self.dst_pros_file1)))
		self.assertEqual(int(pros_file2_mt),int(os.path.getmtime(self.dst_pros_file2)))
		self.assertEqual(int(pros_file3_mt),int(os.path.getmtime(self.dst_pros_file3)))
		self.assertEqual(int(cons_file1_mt),int(os.path.getmtime(self.dst_cons_file1)))
		self.assertEqual(int(cons_file2_mt),int(os.path.getmtime(self.dst_cons_file2)))
		self.assertEqual(int(redundant_1_mt),int(os.path.getmtime(self.dst_redundant_1)))
		self.assertEqual(int(redundant_2_mt),int(os.path.getmtime(self.dst_redundant_2)))








	def test_push_c_crossfs_raise(self):
		class Sakura(Tree):

			bough = self.EASY_SET_BOUGH
			class loggy(LibraryContrib):

				handler		= self.EASY_HANDLER_12
				init_name	= "crossfs_raise"
				init_level	= 10

			@GrowingPeel
			class grow(Transmutable):

				# Actual "LeafPush" imitation with different copy function.
				# The implementation to be checked against original hagrid.planting.leafs.LeafPush!
				def __call__(self, origin :Tree, sprout :str, branch :Path, leaf :Path, bough :Path):

					target	= bough.joinpath(leaf.name)
					shoot	= target.parent
					state	= target.stat().st_mtime if target.is_file() else 0


					if	leaf.is_file():
						if	not shoot.is_dir() : shoot.mkdir(parents=True)

						try:
							LeafPushCases.crossfs_copy(leaf, target, follow_symlinks=False)

							if	target.is_file() and leaf.stat().st_mtime == target.stat().st_mtime:

								leaf.unlink()
								origin.loggy.info(f"Pushed leaf \"{target}\"")
							else:
								origin.loggy.info(f"Leaf \"{target}\" push failed")


						except:
							if	target.is_file():
								if	state <target.stat().st_mtime:

									origin.loggy.info(f"Leaf \"{target}\" pushing stuck, grown instead")
									return
							raise
					else:	origin.loggy.debug(f"Branch \"{leaf}\" not located")

			class files(Rejuvenation):	pass

			@fssprout(self.EASY_SET_SPROUT)
			class sync(Flourish):		pass


		sleep(1.5)
		if	os.path.isdir(self.EASY_SET_BOUGH): rmtree(self.EASY_SET_BOUGH)
		os.makedirs(self.EASY_SET_BOUGH, exist_ok=True)


		self.assertTrue(os.path.isfile(self.file1))
		self.assertTrue(os.path.isdir(self.pros_folder))
		self.assertTrue(os.path.isfile(self.pros_file1))
		self.assertTrue(os.path.isfile(self.pros_file2))
		self.assertTrue(os.path.isfile(self.pros_file3))
		self.assertTrue(os.path.isdir(self.cons_folder))
		self.assertTrue(os.path.isfile(self.cons_file1))
		self.assertTrue(os.path.isfile(self.cons_file2))
		self.assertTrue(os.path.isdir(self.redundant_1_folder))
		self.assertTrue(os.path.isdir(self.redundant_2_folder))
		self.assertTrue(os.path.isfile(self.redundant_1))
		self.assertTrue(os.path.isfile(self.redundant_2))


		self.assertFalse(os.path.isfile(self.dst_file1))
		self.assertFalse(os.path.isdir(self.dst_pros_folder))
		self.assertFalse(os.path.isfile(self.dst_pros_file1))
		self.assertFalse(os.path.isfile(self.dst_pros_file2))
		self.assertFalse(os.path.isfile(self.dst_pros_file3))
		self.assertFalse(os.path.isdir(self.dst_cons_folder))
		self.assertFalse(os.path.isfile(self.dst_cons_file1))
		self.assertFalse(os.path.isfile(self.dst_cons_file2))
		self.assertFalse(os.path.isdir(self.dst_redundant_1_folder))
		self.assertFalse(os.path.isdir(self.dst_redundant_2_folder))
		self.assertFalse(os.path.isfile(self.dst_redundant_1))
		self.assertFalse(os.path.isfile(self.dst_redundant_2))


		with self.assertLogs("crossfs_raise", 10) as case_loggy:

			self.test_case = Sakura()
			self.test_case.sync()


		self.no_loggy_levels(case_loggy.output, 30,40,50)


		self.assertTrue(os.path.isfile(self.file1))
		self.assertTrue(os.path.isdir(self.pros_folder))
		self.assertTrue(os.path.isfile(self.pros_file1))
		self.assertTrue(os.path.isfile(self.pros_file2))
		self.assertTrue(os.path.isfile(self.pros_file3))
		self.assertTrue(os.path.isdir(self.cons_folder))
		self.assertTrue(os.path.isfile(self.cons_file1))
		self.assertTrue(os.path.isfile(self.cons_file2))
		self.assertTrue(os.path.isdir(self.redundant_1_folder))
		self.assertTrue(os.path.isdir(self.redundant_2_folder))
		self.assertTrue(os.path.isfile(self.redundant_1))
		self.assertTrue(os.path.isfile(self.redundant_2))


		self.assertTrue(os.path.isfile(self.dst_file1))
		self.assertTrue(os.path.isdir(self.dst_pros_folder))
		self.assertTrue(os.path.isfile(self.dst_pros_file1))
		self.assertTrue(os.path.isfile(self.dst_pros_file2))
		self.assertTrue(os.path.isfile(self.dst_pros_file3))
		self.assertTrue(os.path.isdir(self.dst_cons_folder))
		self.assertTrue(os.path.isfile(self.dst_cons_file1))
		self.assertTrue(os.path.isfile(self.dst_cons_file2))
		self.assertTrue(os.path.isdir(self.dst_redundant_1_folder))
		self.assertTrue(os.path.isdir(self.dst_redundant_2_folder))
		self.assertTrue(os.path.isfile(self.dst_redundant_1))
		self.assertTrue(os.path.isfile(self.dst_redundant_2))


		self.assertIn(

			f"INFO:crossfs_raise:Leaf \"{self.dst_file1}\" pushing stuck, grown instead",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:crossfs_raise:Leaf \"{self.dst_pros_file1}\" pushing stuck, grown instead",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:crossfs_raise:Leaf \"{self.dst_pros_file2}\" pushing stuck, grown instead",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:crossfs_raise:Leaf \"{self.dst_pros_file3}\" pushing stuck, grown instead",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:crossfs_raise:Leaf \"{self.dst_cons_file1}\" pushing stuck, grown instead",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:crossfs_raise:Leaf \"{self.dst_cons_file2}\" pushing stuck, grown instead",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:crossfs_raise:Leaf \"{self.dst_redundant_1}\" pushing stuck, grown instead",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:crossfs_raise:Leaf \"{self.dst_redundant_2}\" pushing stuck, grown instead",
			case_loggy.output
		)


		self.assertLess(

			int(os.path.getmtime(self.file1)),
			int(os.path.getmtime(self.dst_file1))
		)
		self.assertLess(

			int(os.path.getmtime(self.pros_file1)),
			int(os.path.getmtime(self.dst_pros_file1))
		)
		self.assertLess(

			int(os.path.getmtime(self.pros_file2)),
			int(os.path.getmtime(self.dst_pros_file2))
		)
		self.assertLess(

			int(os.path.getmtime(self.pros_file3)),
			int(os.path.getmtime(self.dst_pros_file3))
		)
		self.assertLess(

			int(os.path.getmtime(self.cons_file1)),
			int(os.path.getmtime(self.dst_cons_file1))
		)
		self.assertLess(

			int(os.path.getmtime(self.cons_file2)),
			int(os.path.getmtime(self.dst_cons_file2))
		)
		self.assertLess(

			int(os.path.getmtime(self.redundant_1)),
			int(os.path.getmtime(self.dst_redundant_1))
		)
		self.assertLess(

			int(os.path.getmtime(self.redundant_2)),
			int(os.path.getmtime(self.dst_redundant_2))
		)








	def test_push_d_only_raise(self):
		class Sakura(Tree):

			bough = self.EASY_SET_BOUGH
			class loggy(LibraryContrib):

				handler		= self.EASY_HANDLER_12
				init_name	= "push_only_raise"
				init_level	= 10

			@GrowingPeel
			class grow(Transmutable):

				# Actual "LeafPush" imitation with different copy function.
				# The implementation to be checked against original hagrid.planting.leafs.LeafPush!
				def __call__(self, origin :Tree, sprout :str, branch :Path, leaf :Path, bough :Path):

					target	= bough.joinpath(leaf.name)
					shoot	= target.parent
					state	= target.stat().st_mtime if target.is_file() else 0


					if	leaf.is_file():
						if	not shoot.is_dir() : shoot.mkdir(parents=True)

						try:
							LeafPushCases.dontlike_copy(leaf, target, follow_symlinks=False)

							if	target.is_file() and leaf.stat().st_mtime == target.stat().st_mtime:

								leaf.unlink()
								origin.loggy.info(f"Pushed leaf \"{target}\"")
							else:
								origin.loggy.info(f"Leaf \"{target}\" push failed")


						except:
							if	target.is_file():
								if	state <target.stat().st_mtime:

									origin.loggy.info(f"Leaf \"{target}\" pushing stuck, grown instead")
									return
							raise
					else:	origin.loggy.debug(f"Branch \"{leaf}\" not located")

			class files(Rejuvenation):	pass

			@fssprout(self.EASY_SET_SPROUT)
			class sync(Flourish):		pass


		if	os.path.isdir(self.EASY_SET_BOUGH): rmtree(self.EASY_SET_BOUGH)
		os.makedirs(self.EASY_SET_BOUGH, exist_ok=True)


		self.assertTrue(os.path.isfile(self.file1))
		self.assertTrue(os.path.isdir(self.pros_folder))
		self.assertTrue(os.path.isfile(self.pros_file1))
		self.assertTrue(os.path.isfile(self.pros_file2))
		self.assertTrue(os.path.isfile(self.pros_file3))
		self.assertTrue(os.path.isdir(self.cons_folder))
		self.assertTrue(os.path.isfile(self.cons_file1))
		self.assertTrue(os.path.isfile(self.cons_file2))
		self.assertTrue(os.path.isdir(self.redundant_1_folder))
		self.assertTrue(os.path.isdir(self.redundant_2_folder))
		self.assertTrue(os.path.isfile(self.redundant_1))
		self.assertTrue(os.path.isfile(self.redundant_2))


		self.assertFalse(os.path.isfile(self.dst_file1))
		self.assertFalse(os.path.isdir(self.dst_pros_folder))
		self.assertFalse(os.path.isfile(self.dst_pros_file1))
		self.assertFalse(os.path.isfile(self.dst_pros_file2))
		self.assertFalse(os.path.isfile(self.dst_pros_file3))
		self.assertFalse(os.path.isdir(self.dst_cons_folder))
		self.assertFalse(os.path.isfile(self.dst_cons_file1))
		self.assertFalse(os.path.isfile(self.dst_cons_file2))
		self.assertFalse(os.path.isdir(self.dst_redundant_1_folder))
		self.assertFalse(os.path.isdir(self.dst_redundant_2_folder))
		self.assertFalse(os.path.isfile(self.dst_redundant_1))
		self.assertFalse(os.path.isfile(self.dst_redundant_2))


		with self.assertLogs("push_only_raise", 10) as case_loggy:

			self.test_case = Sakura()
			self.test_case.sync()


		self.assertTrue(os.path.isfile(self.file1))
		self.assertTrue(os.path.isdir(self.pros_folder))
		self.assertTrue(os.path.isfile(self.pros_file1))
		self.assertTrue(os.path.isfile(self.pros_file2))
		self.assertTrue(os.path.isfile(self.pros_file3))
		self.assertTrue(os.path.isdir(self.cons_folder))
		self.assertTrue(os.path.isfile(self.cons_file1))
		self.assertTrue(os.path.isfile(self.cons_file2))
		self.assertTrue(os.path.isdir(self.redundant_1_folder))
		self.assertTrue(os.path.isdir(self.redundant_2_folder))
		self.assertTrue(os.path.isfile(self.redundant_1))
		self.assertTrue(os.path.isfile(self.redundant_2))


		self.assertFalse(os.path.isfile(self.dst_file1))
		self.assertTrue(os.path.isdir(self.dst_pros_folder))
		self.assertFalse(os.path.isfile(self.dst_pros_file1))
		self.assertFalse(os.path.isfile(self.dst_pros_file2))
		self.assertFalse(os.path.isfile(self.dst_pros_file3))
		self.assertTrue(os.path.isdir(self.dst_cons_folder))
		self.assertFalse(os.path.isfile(self.dst_cons_file1))
		self.assertFalse(os.path.isfile(self.dst_cons_file2))
		self.assertTrue(os.path.isdir(self.dst_redundant_1_folder))
		self.assertTrue(os.path.isdir(self.dst_redundant_2_folder))
		self.assertFalse(os.path.isfile(self.dst_redundant_1))
		self.assertFalse(os.path.isfile(self.dst_redundant_2))


		self.assertIn(

			f"WARNING:push_only_raise:Failed to rejuve leaf \"{self.file1}\" "
			f"due to ValueError: I don't like \"{self.file1}\"",
			case_loggy.output
		)
		self.assertIn(

			f"WARNING:push_only_raise:Failed to rejuve leaf \"{self.pros_file1}\" "
			f"due to ValueError: I don't like \"{self.pros_file1}\"",
			case_loggy.output
		)
		self.assertIn(

			f"WARNING:push_only_raise:Failed to rejuve leaf \"{self.pros_file2}\" "
			f"due to ValueError: I don't like \"{self.pros_file2}\"",
			case_loggy.output
		)
		self.assertIn(

			f"WARNING:push_only_raise:Failed to rejuve leaf \"{self.pros_file3}\" "
			f"due to ValueError: I don't like \"{self.pros_file3}\"",
			case_loggy.output
		)
		self.assertIn(

			f"WARNING:push_only_raise:Failed to rejuve leaf \"{self.cons_file1}\" "
			f"due to ValueError: I don't like \"{self.cons_file1}\"",
			case_loggy.output
		)
		self.assertIn(

			f"WARNING:push_only_raise:Failed to rejuve leaf \"{self.cons_file2}\" "
			f"due to ValueError: I don't like \"{self.cons_file2}\"",
			case_loggy.output
		)
		self.assertIn(

			f"WARNING:push_only_raise:Failed to rejuve leaf \"{self.redundant_1}\" "
			f"due to ValueError: I don't like \"{self.redundant_1}\"",
			case_loggy.output
		)
		self.assertIn(

			f"WARNING:push_only_raise:Failed to rejuve leaf \"{self.redundant_2}\" "
			f"due to ValueError: I don't like \"{self.redundant_2}\"",
			case_loggy.output
		)








	def test_push_e_empty(self):
		class Sakura(Tree):

			bough = self.EASY_SET_BOUGH
			class loggy(LibraryContrib):

				handler		= self.EASY_HANDLER_12
				init_name	= "push_empty"
				init_level	= 10

			@GrowingPeel
			class grow(Transmutable):

				# Actual "LeafPush" imitation with different copy function.
				# The implementation to be checked against original hagrid.planting.leafs.LeafPush!
				def __call__(self, origin :Tree, sprout :str, branch :Path, leaf :Path, bough :Path):

					target	= bough.joinpath(leaf.name)
					shoot	= target.parent
					state	= target.stat().st_mtime if target.is_file() else 0


					if	leaf.is_file():
						if	not shoot.is_dir() : shoot.mkdir(parents=True)

						try:
							LeafPushCases.empty_copy(leaf, target, follow_symlinks=False)

							if	target.is_file() and leaf.stat().st_mtime == target.stat().st_mtime:

								leaf.unlink()
								origin.loggy.info(f"Pushed leaf \"{target}\"")
							else:
								origin.loggy.info(f"Leaf \"{target}\" push failed")


						except:
							if	target.is_file():
								if	state <target.stat().st_mtime:

									origin.loggy.info(f"Leaf \"{target}\" pushing stuck, grown instead")
									return
							raise
					else:	origin.loggy.debug(f"Branch \"{leaf}\" not located")

			class files(Rejuvenation):	pass

			@fssprout(self.EASY_SET_SPROUT)
			class sync(Flourish):		pass


		if	os.path.isdir(self.EASY_SET_BOUGH): rmtree(self.EASY_SET_BOUGH)
		os.makedirs(self.EASY_SET_BOUGH, exist_ok=True)


		self.assertTrue(os.path.isfile(self.file1))
		self.assertTrue(os.path.isdir(self.pros_folder))
		self.assertTrue(os.path.isfile(self.pros_file1))
		self.assertTrue(os.path.isfile(self.pros_file2))
		self.assertTrue(os.path.isfile(self.pros_file3))
		self.assertTrue(os.path.isdir(self.cons_folder))
		self.assertTrue(os.path.isfile(self.cons_file1))
		self.assertTrue(os.path.isfile(self.cons_file2))
		self.assertTrue(os.path.isdir(self.redundant_1_folder))
		self.assertTrue(os.path.isdir(self.redundant_2_folder))
		self.assertTrue(os.path.isfile(self.redundant_1))
		self.assertTrue(os.path.isfile(self.redundant_2))


		self.assertFalse(os.path.isfile(self.dst_file1))
		self.assertFalse(os.path.isdir(self.dst_pros_folder))
		self.assertFalse(os.path.isfile(self.dst_pros_file1))
		self.assertFalse(os.path.isfile(self.dst_pros_file2))
		self.assertFalse(os.path.isfile(self.dst_pros_file3))
		self.assertFalse(os.path.isdir(self.dst_cons_folder))
		self.assertFalse(os.path.isfile(self.dst_cons_file1))
		self.assertFalse(os.path.isfile(self.dst_cons_file2))
		self.assertFalse(os.path.isdir(self.dst_redundant_1_folder))
		self.assertFalse(os.path.isdir(self.dst_redundant_2_folder))
		self.assertFalse(os.path.isfile(self.dst_redundant_1))
		self.assertFalse(os.path.isfile(self.dst_redundant_2))


		with self.assertLogs("push_empty", 10) as case_loggy:

			self.test_case = Sakura()
			self.test_case.sync()


		self.assertTrue(os.path.isfile(self.file1))
		self.assertTrue(os.path.isdir(self.pros_folder))
		self.assertTrue(os.path.isfile(self.pros_file1))
		self.assertTrue(os.path.isfile(self.pros_file2))
		self.assertTrue(os.path.isfile(self.pros_file3))
		self.assertTrue(os.path.isdir(self.cons_folder))
		self.assertTrue(os.path.isfile(self.cons_file1))
		self.assertTrue(os.path.isfile(self.cons_file2))
		self.assertTrue(os.path.isdir(self.redundant_1_folder))
		self.assertTrue(os.path.isdir(self.redundant_2_folder))
		self.assertTrue(os.path.isfile(self.redundant_1))
		self.assertTrue(os.path.isfile(self.redundant_2))


		self.assertFalse(os.path.isfile(self.dst_file1))
		self.assertTrue(os.path.isdir(self.dst_pros_folder))
		self.assertFalse(os.path.isfile(self.dst_pros_file1))
		self.assertFalse(os.path.isfile(self.dst_pros_file2))
		self.assertFalse(os.path.isfile(self.dst_pros_file3))
		self.assertTrue(os.path.isdir(self.dst_cons_folder))
		self.assertFalse(os.path.isfile(self.dst_cons_file1))
		self.assertFalse(os.path.isfile(self.dst_cons_file2))
		self.assertTrue(os.path.isdir(self.dst_redundant_1_folder))
		self.assertTrue(os.path.isdir(self.dst_redundant_2_folder))
		self.assertFalse(os.path.isfile(self.dst_redundant_1))
		self.assertFalse(os.path.isfile(self.dst_redundant_2))


		self.assertIn(

			f"INFO:push_empty:Leaf \"{self.dst_file1}\" push failed",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_empty:Leaf \"{self.dst_pros_file1}\" push failed",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_empty:Leaf \"{self.dst_pros_file2}\" push failed",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_empty:Leaf \"{self.dst_pros_file3}\" push failed",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_empty:Leaf \"{self.dst_cons_file1}\" push failed",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_empty:Leaf \"{self.dst_cons_file2}\" push failed",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_empty:Leaf \"{self.dst_redundant_1}\" push failed",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:push_empty:Leaf \"{self.dst_redundant_2}\" push failed",
			case_loggy.output
		)








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	os
import	re
import	unittest
from	pygwarts.magical.philosophers_stone	import Transmutable
from	pygwarts.tests.irma					import IrmaTestCase
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.irma.access				import LibraryAccess
from	pygwarts.irma.access.volume			import LibraryVolume








class AccessVolumeCase(IrmaTestCase):

	"""
		Testing for fields and edge cases
	"""

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.ACCESS_VOLUME): os.remove(cls.ACCESS_VOLUME)

	@classmethod
	def setUpClass(cls):

		super().setUpClass()
		cls.make_loggy_file(cls, cls.ACCESS_VOLUME)
		cls.some_location = os.path.join(os.path.dirname(cls.ACCESS_VOLUME), "some_location")

	def setUp(self):

		if	os.path.isfile(self.some_location):	os.remove(self.some_location)
		if	os.path.isdir(self.some_location):	os.rmdir(self.some_location)


	def test_no_upper_init(self):

		class LonelyVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_VOLUME
				init_name	= "no_upper_init"
				init_level	= 10

		with self.assertLogs("no_upper_init", 10) as case_loggy : self.test_case = LonelyVolume()
		self.assertIn(
			f"INFO:no_upper_init:Volume {self.test_case} not assigned to any library", case_loggy.output
		)


	def test_LibraryAccess_init(self):
		class LonelyLibrary(LibraryAccess):

			class LonelyVolume(LibraryVolume):	pass
			class loggy(LibraryContrib):

				handler		= self.ACCESS_VOLUME
				init_name	= "LibraryAccess_init"
				init_level	= 10

		with self.assertLogs("LibraryAccess_init", 10) as case_loggy : self.test_case = LonelyLibrary()
		self.assertIn(f"DEBUG:LibraryAccess_init:Assigned to {self.test_case}", case_loggy.output)








	def test_get_range_valids(self):
		class LonelyVolume(LibraryVolume):	pass

		self.test_case = LonelyVolume()
		self.assertEqual(self.test_case.get_range("0420"), re.compile("0420"))
		self.test_case.inrange = "01/01"
		self.assertEqual(self.test_case.get_range(), re.compile("01/01"))


	def test_get_range_invalids(self):

		class LonelyVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_VOLUME
				init_name	= "get_range_invalids"
				init_level	= 10

		self.test_case = LonelyVolume()
		for invalid in (

			42, .69, None, Transmutable, print, True, False, ...,
			[ "0420" ],( "0420", ),{ "0420" },{ "inrange": "0420" }
		):
			with self.subTest(invalid=invalid):
				with self.assertLogs("get_range_invalids", 10) as case_loggy:

					self.assertIsNone(self.test_case.get_range(invalid))
					self.test_case.inrange = invalid
					self.assertIsNone(self.test_case.get_range())

				self.assertIn(

					f"INFO:get_range_invalids:Valid {self.test_case} range not provided",
					case_loggy.output
				)
				self.assertEqual(
					case_loggy.output.count(

						f"INFO:get_range_invalids:Valid {self.test_case} range not provided"
					),	2
				)








	def test_is_located(self):

		class LonelyVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_VOLUME
				init_name	= "is_located"
				init_level	= 10

		self.test_case = LonelyVolume()
		with self.assertLogs("is_located", 10) as case_loggy:
			self.assertIsNone(self.test_case.is_located())

		self.assertIn("DEBUG:is_located:Location was not provided", case_loggy.output)

		self.test_case.location = self.some_location
		with self.assertLogs("is_located", 10) as case_loggy:
			self.assertIsNone(self.test_case.is_located())

		self.assertIn(f"DEBUG:is_located:Location \"{self.some_location}\" is invalid", case_loggy.output)

		os.mkdir(self.some_location)
		with self.assertLogs("is_located", 10) as case_loggy:
			self.assertIsNone(self.test_case.is_located())

		self.assertIn(f"DEBUG:is_located:Location \"{self.some_location}\" is invalid", case_loggy.output)

		os.rmdir(self.some_location)
		self.fmake(self.some_location)
		with self.assertLogs("is_located", 10) as case_loggy:
			self.assertEqual(self.test_case.is_located(), self.some_location)

		self.assertIn(f"DEBUG:is_located:Accessible \"{self.some_location}\" located", case_loggy.output)

		os.remove(self.some_location)
		self.assertFalse(os.path.isfile(self.some_location))
		self.assertFalse(os.path.isdir(self.some_location))

		#
		# Bonus for normal OS. location might be set to any existent but lack of r-bit file
		#
		#self.test_case.location = "/swapfile"
		#with self.assertLogs("is_located", 10) as case_loggy:
		#	self.assertIsNone(self.test_case.is_located())

		#self.assertIn(
		#	f"DEBUG:is_located:Location \"/swapfile\" is not accessible", case_loggy.output
		#)








	def test_read_generator(self):
		class LonelyVolume(LibraryVolume):

			location = self.some_location
			class loggy(LibraryContrib):

				handler		= self.ACCESS_VOLUME
				init_name	= "read_generator"
				init_level	= 10

		self.fmake(self.some_location, "OOH\nEEH\nOOH\nAH\nAH\nTING\nTANG\nWALLA\nWALLA\nBING\nBANG")
		self.test_case = LonelyVolume()

		with self.assertLogs("read_generator", 10) as case_loggy:
			self.assertEqual(

				list(self.test_case.g_reading()),
				[ "OOH", "EEH", "OOH", "AH", "AH", "TING", "TANG", "WALLA", "WALLA", "BING", "BANG" ]
			)
		self.assertIn(
			f"DEBUG:read_generator:Location \"{self.some_location}\" access granted", case_loggy.output
		)
		self.assertEqual(self.test_case.is_located(), self.some_location)

		with self.assertLogs("read_generator", 10) as case_loggy:

			lines = []
			for line in self.test_case.g_reading() : lines.append(line)
			self.assertEqual(

				lines,
				[ "OOH", "EEH", "OOH", "AH", "AH", "TING", "TANG", "WALLA", "WALLA", "BING", "BANG" ]
			)
		self.assertIn(
			f"DEBUG:read_generator:Location \"{self.some_location}\" access granted", case_loggy.output
		)
		self.assertEqual(self.test_case.is_located(), self.some_location)




	def test_read_generator_no_file(self):
		class LonelyVolume(LibraryVolume):

			location = self.some_location
			class loggy(LibraryContrib):

				handler		= self.ACCESS_VOLUME
				init_name	= "read_generator_no_file"
				init_level	= 10

		self.test_case = LonelyVolume()
		self.assertFalse(os.path.isfile(self.some_location))
		self.assertFalse(os.path.isdir(self.some_location))

		with self.assertLogs("read_generator_no_file", 10) as case_loggy:
			lines = []
			for line in self.test_case.g_reading() : lines.append(line)

		self.assertIn(
			f"DEBUG:read_generator_no_file:Location \"{self.some_location}\" is invalid", case_loggy.output
		)
		self.assertIsNone(self.test_case.is_located())




	def test_read_generator_break_outer(self):
		class LonelyVolume(LibraryVolume):

			location = self.some_location
			class loggy(LibraryContrib):

				handler		= self.ACCESS_VOLUME
				init_name	= "read_generator_break_outer"
				init_level	= 10

		self.fmake(self.some_location, "OOH\nEEH\nOOH\nAH\nAH\nTING\nTANG\nWALLA\nWALLA\nBING\nBANG")
		self.test_case = LonelyVolume()

		with self.assertLogs("read_generator_break_outer", 10) as case_loggy:

			lines = []
			for line in self.test_case.g_reading():

				if	line == "AH" : break
				lines.append(line)

			self.assertEqual(lines,[ "OOH", "EEH", "OOH" ])
		self.assertIn(

			f"DEBUG:read_generator_break_outer:Location \"{self.some_location}\" access granted",
			case_loggy.output
		)
		self.assertEqual(self.test_case.is_located(), self.some_location)








	def test_triggering_generator(self):
		class LonelyVolume(LibraryVolume):

			location = self.some_location
			class loggy(LibraryContrib):

				handler		= self.ACCESS_VOLUME
				init_name	= "triggering_generator"
				init_level	= 10

		line = "OOH EEH OOH AH AH TING TANG WALLA WALLA BING BANG"
		triggers = re.compile("(OOH EEH)|(ooh ah ah)|(WALLA WALLA)")
		self.test_case = LonelyVolume()

		with self.assertLogs("triggering_generator", 10) as case_loggy:
			self.assertEqual(

				list(self.test_case.g_triggering(line, triggers)),
				[ "OOH EEH","WALLA WALLA" ]
			)

		self.assertIn("DEBUG:triggering_generator:Triggered by \"OOH EEH\"", case_loggy.output)
		self.assertIn("DEBUG:triggering_generator:Triggered by \"WALLA WALLA\"", case_loggy.output)


		with self.assertLogs("triggering_generator", 10) as case_loggy:

			responses = []
			for response in self.test_case.g_triggering(line, triggers) : responses.append(response)
			self.assertEqual(responses,[ "OOH EEH","WALLA WALLA" ])

		self.assertIn("DEBUG:triggering_generator:Triggered by \"OOH EEH\"", case_loggy.output)
		self.assertIn("DEBUG:triggering_generator:Triggered by \"WALLA WALLA\"", case_loggy.output)




	def test_triggering_generator_bad_triggers(self):
		class LonelyVolume(LibraryVolume):

			location = self.some_location
			class loggy(LibraryContrib):

				handler		= self.ACCESS_VOLUME
				init_name	= "triggering_generator_bad_triggers"
				init_level	= 10

		self.test_case = LonelyVolume()
		for bad_triggers in (

			"triggers", 42, .69, True, False, None, ..., print, Transmutable,
			( "trigger #1", "trigger #2", "trigger #3" ),
			{ "trigger #1", "trigger #2", "trigger #3" },
			{ 1: "trigger #1", 2: "trigger #2", 3: "trigger #3" },
		):
			with self.subTest(triggers=bad_triggers):
				with self.assertLogs("triggering_generator_bad_triggers", 10) as case_loggy:
					self.assertEqual(
						list(self.test_case.g_triggering("no need line", bad_triggers)),[]
					)
				self.assertIn(

					f"DEBUG:triggering_generator_bad_triggers:Invalid triggers \"{bad_triggers}\"",
					case_loggy.output
				)




	def test_triggering_generator_bad_line(self):
		class LonelyVolume(LibraryVolume):

			location = self.some_location
			class loggy(LibraryContrib):

				handler		= self.ACCESS_VOLUME
				init_name	= "triggering_generator_bad_line"
				init_level	= 10

		self.test_case = LonelyVolume()
		triggers = re.compile("(no)|(need)|(line)")
		for bad_line in (

			42, .69, True, False, None, ..., print, Transmutable,
			[ "no need line" ],
			( "no need line", ),
			{ "no need line" },
			{ "line": "no need line" },
		):
			with self.subTest(line=bad_line):
				with self.assertLogs("triggering_generator_bad_line", 10) as case_loggy:
					self.assertEqual(
						list(self.test_case.g_triggering(bad_line, triggers)),[]
					)
				self.assertIn(

					f"DEBUG:triggering_generator_bad_line:Invalid line \"{bad_line}\"",
					case_loggy.output
				)




	def test_triggering_generator_no_response(self):
		class LonelyVolume(LibraryVolume):

			location = self.some_location
			class loggy(LibraryContrib):

				handler		= self.ACCESS_VOLUME
				init_name	= "triggering_generator_no_response"
				init_level	= 10

		line = "OOH EEH OOH AH AH TING TANG WALLA WALLA BING BANG"
		triggers = re.compile("(ooh EEH)|(ooh AH AH)|(walla WALLA)")

		self.test_case = LonelyVolume()
		self.assertEqual(list(self.test_case.g_triggering(line, triggers)),[])




	def test_triggering_generator_break(self):
		class LonelyVolume(LibraryVolume):

			location = self.some_location
			class loggy(LibraryContrib):

				handler		= self.ACCESS_VOLUME
				init_name	= "triggering_generator_break"
				init_level	= 10

		line = "OOH EEH OOH AH AH TING TANG WALLA WALLA BING BANG"
		triggers = re.compile("(OOH EEH)|(OOH AH AH)|(WALLA WALLA)")
		self.test_case = LonelyVolume()

		with self.assertLogs("triggering_generator_break", 10) as case_loggy:

			responses = []
			for response in self.test_case.g_triggering(line, triggers):

				responses.append(response)
				break

			self.assertEqual(responses,[ "OOH EEH" ])
		self.assertIn("DEBUG:triggering_generator_break:Triggered by \"OOH EEH\"", case_loggy.output)




	def test_triggering_generator_single_trigger(self):
		class LonelyVolume(LibraryVolume):

			location = self.some_location
			class loggy(LibraryContrib):

				handler		= self.ACCESS_VOLUME
				init_name	= "triggering_generator_single_trigger"
				init_level	= 10

		line = "OOH EEH OOH AH AH TING TANG WALLA WALLA BING BANG"
		triggers = re.compile("(OOH EEH)")
		self.test_case = LonelyVolume()

		with self.assertLogs("triggering_generator_single_trigger", 10) as case_loggy:
			self.assertEqual(list(self.test_case.g_triggering(line, triggers)),[ "OOH EEH" ])
		self.assertIn("DEBUG:triggering_generator_single_trigger:Triggered by \"OOH EEH\"",case_loggy.output)








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	os
import	unittest
from	pathlib								import Path
from	time								import sleep
from	shutil								import rmtree
from	pygwarts.magical.philosophers_stone	import Transmutable
from	pygwarts.magical.time_turner		import TimeTurner
from	pygwarts.tests.hagrid				import HagridTestCase
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.hagrid.planting.peeks		import DraftPeek








class DraftPeekCases(HagridTestCase):

	"""
		Ensures peeks algorithm works as intended
	"""

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.DRAFT_PEEK_HANDLER): os.remove(cls.DRAFT_PEEK_HANDLER)

		if	os.path.isdir(cls.INIT_SET_SPROUT):		rmtree(cls.INIT_SET_SPROUT)
		if	os.path.isdir(cls.INIT_SET_BOUGH_1):	rmtree(cls.INIT_SET_BOUGH_1)

	@classmethod
	def setUpClass(cls):

		cls.peekable1 = "peekable1"
		cls.make_loggy_file(cls, cls.DRAFT_PEEK_HANDLER)
		if not os.path.isdir(cls.INIT_SET_BOUGH_1): os.makedirs(cls.INIT_SET_BOUGH_1)

		cls.peekable_sprout1	= os.path.join(cls.INIT_SET_SPROUT, "peekable1")
		cls.peekable_bough1		= os.path.join(cls.INIT_SET_BOUGH_1, "peekable1")
		cls.peekable_bough2		= os.path.join(cls.INIT_SET_BOUGH_2, "peekable1")
		cls.peekable_bough3		= os.path.join(cls.INIT_SET_BOUGH_3, "peekable1")




	def test_cached_draft_renew_1(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.DRAFT_PEEK_HANDLER
				init_name	= "cached_renew_1"
				init_level	= 10

			@DraftPeek()
			class grow(Transmutable):
				def __call__(*args):	pass


		self.test_case = Sakura()
		self.fmake(self.peekable_sprout1)
		first_peek = int(os.path.getmtime(self.peekable_sprout1))
		if	os.path.isfile(self.peekable_bough1): os.remove(self.peekable_bough1)
		if	os.path.isfile(self.peekable_bough2): os.remove(self.peekable_bough2)
		if	os.path.isfile(self.peekable_bough3): os.remove(self.peekable_bough3)


		with self.assertLogs("cached_renew_1", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:cached_renew_1:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:cached_renew_1:Flourish stopped cause draft peek only for sprigs renew",
				case_loggy.output
			)


		with self.assertLogs("cached_renew_1", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:cached_renew_1:Cached draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:cached_renew_1:Flourish stopped cause draft peek only for sprigs renew",
				case_loggy.output
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)


		with self.assertLogs("cached_renew_1", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:cached_renew_1:Cached draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:cached_renew_1:Flourish stopped cause draft peek only for sprigs renew",
				case_loggy.output
			)








	def test_cached_draft_renew_2(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.DRAFT_PEEK_HANDLER
				init_name	= "cached_renew_2"
				init_level	= 10

			@DraftPeek()
			class grow(Transmutable):
				def __call__(*args):	pass


		self.test_case = Sakura()
		self.fmake(self.peekable_bough1)
		bough1_peek = int(os.path.getmtime(self.peekable_bough1))
		self.fmake(self.peekable_bough2)
		bough2_peek = int(os.path.getmtime(self.peekable_bough2))
		self.fmake(self.peekable_bough3)
		bough3_peek = int(os.path.getmtime(self.peekable_bough3))

		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		first_peek = int(os.path.getmtime(self.peekable_sprout1))


		with self.assertLogs("cached_renew_2", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:cached_renew_2:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_renew_2:Draft peek for \"{self.peekable_bough1}\" old ring: {bough1_peek}",
				case_loggy.output
			)


		with self.assertLogs("cached_renew_2", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:cached_renew_2:Cached draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_renew_2:Draft peek for \"{self.peekable_bough2}\" old ring: {bough2_peek}",
				case_loggy.output
			)




		sleep(1.1)
		self.fmake(self.peekable_sprout1)


		with self.assertLogs("cached_renew_2", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:cached_renew_2:Cached draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_renew_2:Draft peek for \"{self.peekable_bough3}\" old ring: {bough3_peek}",
				case_loggy.output
			)








	def test_cached_draft_1(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.DRAFT_PEEK_HANDLER
				init_name	= "cached_1"
				init_level	= 10

			@DraftPeek(renew=False)
			class grow(Transmutable):
				def __call__(*args):	pass


		self.test_case = Sakura()
		self.fmake(self.peekable_sprout1)
		first_peek = int(os.path.getmtime(self.peekable_sprout1))
		if	os.path.isfile(self.peekable_bough1): os.remove(self.peekable_bough1)
		if	os.path.isfile(self.peekable_bough2): os.remove(self.peekable_bough2)
		if	os.path.isfile(self.peekable_bough3): os.remove(self.peekable_bough3)


		with self.assertLogs("cached_1", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:cached_1:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_1:Draft peek for \"{self.peekable_bough1}\" old ring: 0",
				case_loggy.output
			)


		with self.assertLogs("cached_1", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:cached_1:Cached draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_1:Draft peek for \"{self.peekable_bough2}\" old ring: 0",
				case_loggy.output
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)


		with self.assertLogs("cached_1", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:cached_1:Cached draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_1:Draft peek for \"{self.peekable_bough3}\" old ring: 0",
				case_loggy.output
			)








	def test_cached_draft_2(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.DRAFT_PEEK_HANDLER
				init_name	= "cached_2"
				init_level	= 10

			@DraftPeek(renew=False)
			class grow(Transmutable):
				def __call__(*args):	pass


		self.test_case = Sakura()
		self.fmake(self.peekable_bough1)
		bough1_peek = int(os.path.getmtime(self.peekable_bough1))
		self.fmake(self.peekable_bough2)
		bough2_peek = int(os.path.getmtime(self.peekable_bough2))
		self.fmake(self.peekable_bough3)
		bough3_peek = int(os.path.getmtime(self.peekable_bough3))

		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		first_peek = int(os.path.getmtime(self.peekable_sprout1))


		with self.assertLogs("cached_2", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:cached_2:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_2:Draft peek for \"{self.peekable_bough1}\" old ring: {bough1_peek}",
				case_loggy.output
			)


		with self.assertLogs("cached_2", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:cached_2:Cached draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_2:Draft peek for \"{self.peekable_bough2}\" old ring: {bough2_peek}",
				case_loggy.output
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)


		with self.assertLogs("cached_2", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:cached_2:Cached draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_2:Draft peek for \"{self.peekable_bough3}\" old ring: {bough3_peek}",
				case_loggy.output
			)








	def test_cached_draft_picky_1(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.DRAFT_PEEK_HANDLER
				init_name	= "cached_picky_1"
				init_level	= 10

			@DraftPeek(renew=False, picky=True)
			class grow(Transmutable):
				def __call__(*args):	pass


		self.test_case = Sakura()
		self.fmake(self.peekable_sprout1)
		first_peek = os.path.getmtime(self.peekable_sprout1)
		if	os.path.isfile(self.peekable_bough1): os.remove(self.peekable_bough1)
		if	os.path.isfile(self.peekable_bough2): os.remove(self.peekable_bough2)
		if	os.path.isfile(self.peekable_bough3): os.remove(self.peekable_bough3)


		with self.assertLogs("cached_picky_1", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:cached_picky_1:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_picky_1:Draft peek for \"{self.peekable_bough1}\" old ring: 0.0",
				case_loggy.output
			)


		with self.assertLogs("cached_picky_1", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:cached_picky_1:Cached draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_picky_1:Draft peek for \"{self.peekable_bough2}\" old ring: 0.0",
				case_loggy.output
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)


		with self.assertLogs("cached_picky_1", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:cached_picky_1:Cached draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_picky_1:Draft peek for \"{self.peekable_bough3}\" old ring: 0.0",
				case_loggy.output
			)








	def test_cached_draft_picky_2(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.DRAFT_PEEK_HANDLER
				init_name	= "cached_picky_2"
				init_level	= 10

			@DraftPeek(renew=False, picky=True)
			class grow(Transmutable):
				def __call__(*args):	pass


		self.test_case = Sakura()
		self.fmake(self.peekable_bough1)
		bough1_peek = os.path.getmtime(self.peekable_bough1)
		self.fmake(self.peekable_bough2)
		bough2_peek = os.path.getmtime(self.peekable_bough2)
		self.fmake(self.peekable_bough3)
		bough3_peek = os.path.getmtime(self.peekable_bough3)

		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		first_peek = os.path.getmtime(self.peekable_sprout1)


		with self.assertLogs("cached_picky_2", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:cached_picky_2:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_picky_2:Draft peek for \"{self.peekable_bough1}\" old ring: {bough1_peek}",
				case_loggy.output
			)


		with self.assertLogs("cached_picky_2", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:cached_picky_2:Cached draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_picky_2:Draft peek for \"{self.peekable_bough2}\" old ring: {bough2_peek}",
				case_loggy.output
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)


		with self.assertLogs("cached_picky_2", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:cached_picky_2:Cached draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_picky_2:Draft peek for \"{self.peekable_bough3}\" old ring: {bough3_peek}",
				case_loggy.output
			)








	def test_cached_draft_picky_renew_1(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.DRAFT_PEEK_HANDLER
				init_name	= "cached_picky_renew_1"
				init_level	= 10

			@DraftPeek(picky=True)
			class grow(Transmutable):
				def __call__(*args):	pass


		self.test_case = Sakura()
		self.fmake(self.peekable_sprout1)
		first_peek = os.path.getmtime(self.peekable_sprout1)
		if	os.path.isfile(self.peekable_bough1): os.remove(self.peekable_bough1)
		if	os.path.isfile(self.peekable_bough2): os.remove(self.peekable_bough2)
		if	os.path.isfile(self.peekable_bough3): os.remove(self.peekable_bough3)


		with self.assertLogs("cached_picky_renew_1", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:cached_picky_renew_1:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:cached_picky_renew_1:Flourish stopped cause draft peek only for sprigs renew",
				case_loggy.output
			)


		with self.assertLogs("cached_picky_renew_1", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:cached_picky_renew_1:Cached draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:cached_picky_renew_1:Flourish stopped cause draft peek only for sprigs renew",
				case_loggy.output
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)


		with self.assertLogs("cached_picky_renew_1", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:cached_picky_renew_1:Cached draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:cached_picky_renew_1:Flourish stopped cause draft peek only for sprigs renew",
				case_loggy.output
			)








	def test_cached_draft_picky_renew_2(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.DRAFT_PEEK_HANDLER
				init_name	= "cached_picky_renew_2"
				init_level	= 10

			@DraftPeek(picky=True)
			class grow(Transmutable):
				def __call__(*args):	pass


		self.test_case = Sakura()
		self.fmake(self.peekable_bough1)
		bough1_peek = os.path.getmtime(self.peekable_bough1)
		self.fmake(self.peekable_bough2)
		bough2_peek = os.path.getmtime(self.peekable_bough2)
		self.fmake(self.peekable_bough3)
		bough3_peek = os.path.getmtime(self.peekable_bough3)

		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		first_peek = os.path.getmtime(self.peekable_sprout1)


		with self.assertLogs("cached_picky_renew_2", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:cached_picky_renew_2:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_picky_renew_2:Draft peek for \"{self.peekable_bough1}\" "
				f"old ring: {bough1_peek}",
				case_loggy.output
			)


		with self.assertLogs("cached_picky_renew_2", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:cached_picky_renew_2:Cached draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_picky_renew_2:Draft peek for \"{self.peekable_bough2}\" "
				f"old ring: {bough2_peek}",
				case_loggy.output
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)


		with self.assertLogs("cached_picky_renew_2", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:cached_picky_renew_2:Cached draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_picky_renew_2:Draft peek for \"{self.peekable_bough3}\" "
				f"old ring: {bough3_peek}",
				case_loggy.output
			)








	def test_draft_renew_1(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.DRAFT_PEEK_HANDLER
				init_name	= "renew_1"
				init_level	= 10

			@DraftPeek(cache=False)
			class grow(Transmutable):
				def __call__(*args):	pass


		self.test_case = Sakura()
		self.fmake(self.peekable_sprout1)
		first_peek = int(os.path.getmtime(self.peekable_sprout1))
		if	os.path.isfile(self.peekable_bough1): os.remove(self.peekable_bough1)
		if	os.path.isfile(self.peekable_bough2): os.remove(self.peekable_bough2)
		if	os.path.isfile(self.peekable_bough3): os.remove(self.peekable_bough3)


		with self.assertLogs("renew_1", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:renew_1:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:renew_1:Flourish stopped cause draft peek only for sprigs renew",
				case_loggy.output
			)


		with self.assertLogs("renew_1", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:renew_1:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:renew_1:Flourish stopped cause draft peek only for sprigs renew",
				case_loggy.output
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		second_peek = int(os.path.getmtime(self.peekable_sprout1))
		self.assertNotEqual(first_peek, second_peek)


		with self.assertLogs("renew_1", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:renew_1:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{second_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:renew_1:Flourish stopped cause draft peek only for sprigs renew",
				case_loggy.output
			)








	def test_draft_renew_2(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.DRAFT_PEEK_HANDLER
				init_name	= "renew_2"
				init_level	= 10

			@DraftPeek(cache=False)
			class grow(Transmutable):
				def __call__(*args):	pass


		self.test_case = Sakura()
		self.fmake(self.peekable_bough1)
		bough1_peek = int(os.path.getmtime(self.peekable_bough1))
		self.fmake(self.peekable_bough2)
		bough2_peek = int(os.path.getmtime(self.peekable_bough2))
		self.fmake(self.peekable_bough3)
		bough3_peek = int(os.path.getmtime(self.peekable_bough3))

		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		first_peek = int(os.path.getmtime(self.peekable_sprout1))


		with self.assertLogs("renew_2", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:renew_2:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:renew_2:Draft peek for \"{self.peekable_bough1}\" old ring: {bough1_peek}",
				case_loggy.output
			)


		with self.assertLogs("renew_2", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:renew_2:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:renew_2:Draft peek for \"{self.peekable_bough2}\" old ring: {bough2_peek}",
				case_loggy.output
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		second_peek = int(os.path.getmtime(self.peekable_sprout1))
		self.assertNotEqual(first_peek, second_peek)


		with self.assertLogs("renew_2", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:renew_2:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{second_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:renew_2:Draft peek for \"{self.peekable_bough3}\" old ring: {bough3_peek}",
				case_loggy.output
			)








	def test_draft_1(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.DRAFT_PEEK_HANDLER
				init_name	= "1"
				init_level	= 10

			@DraftPeek(renew=False, cache=False)
			class grow(Transmutable):
				def __call__(*args):	pass


		self.test_case = Sakura()
		self.fmake(self.peekable_sprout1)
		first_peek = int(os.path.getmtime(self.peekable_sprout1))
		if	os.path.isfile(self.peekable_bough1): os.remove(self.peekable_bough1)
		if	os.path.isfile(self.peekable_bough2): os.remove(self.peekable_bough2)
		if	os.path.isfile(self.peekable_bough3): os.remove(self.peekable_bough3)


		with self.assertLogs("1", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:1:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:1:Draft peek for \"{self.peekable_bough1}\" old ring: 0",
				case_loggy.output
			)


		with self.assertLogs("1", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:1:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:1:Draft peek for \"{self.peekable_bough2}\" old ring: 0",
				case_loggy.output
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		second_peek = int(os.path.getmtime(self.peekable_sprout1))
		self.assertNotEqual(first_peek, second_peek)


		with self.assertLogs("1", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:1:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{second_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:1:Draft peek for \"{self.peekable_bough3}\" old ring: 0",
				case_loggy.output
			)








	def test_draft_2(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.DRAFT_PEEK_HANDLER
				init_name	= "2"
				init_level	= 10

			@DraftPeek(renew=False, cache=False)
			class grow(Transmutable):
				def __call__(*args):	pass


		self.test_case = Sakura()
		self.fmake(self.peekable_bough1)
		bough1_peek = int(os.path.getmtime(self.peekable_bough1))
		self.fmake(self.peekable_bough2)
		bough2_peek = int(os.path.getmtime(self.peekable_bough2))
		self.fmake(self.peekable_bough3)
		bough3_peek = int(os.path.getmtime(self.peekable_bough3))

		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		first_peek = int(os.path.getmtime(self.peekable_sprout1))


		with self.assertLogs("2", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:2:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:2:Draft peek for \"{self.peekable_bough1}\" old ring: {bough1_peek}",
				case_loggy.output
			)


		with self.assertLogs("2", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:2:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:2:Draft peek for \"{self.peekable_bough2}\" old ring: {bough2_peek}",
				case_loggy.output
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		second_peek = int(os.path.getmtime(self.peekable_sprout1))
		self.assertNotEqual(first_peek, second_peek)


		with self.assertLogs("2", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:2:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{second_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:2:Draft peek for \"{self.peekable_bough3}\" old ring: {bough3_peek}",
				case_loggy.output
			)








	def test_draft_picky_1(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.DRAFT_PEEK_HANDLER
				init_name	= "picky_1"
				init_level	= 10

			@DraftPeek(renew=False, cache=False, picky=True)
			class grow(Transmutable):
				def __call__(*args):	pass


		self.test_case = Sakura()
		self.fmake(self.peekable_sprout1)
		first_peek = os.path.getmtime(self.peekable_sprout1)
		if	os.path.isfile(self.peekable_bough1): os.remove(self.peekable_bough1)
		if	os.path.isfile(self.peekable_bough2): os.remove(self.peekable_bough2)
		if	os.path.isfile(self.peekable_bough3): os.remove(self.peekable_bough3)


		with self.assertLogs("picky_1", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:picky_1:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:picky_1:Draft peek for \"{self.peekable_bough1}\" old ring: 0.0",
				case_loggy.output
			)


		with self.assertLogs("picky_1", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:picky_1:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:picky_1:Draft peek for \"{self.peekable_bough2}\" old ring: 0.0",
				case_loggy.output
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		second_peek = os.path.getmtime(self.peekable_sprout1)
		self.assertNotEqual(first_peek, second_peek)


		with self.assertLogs("picky_1", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:picky_1:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{second_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:picky_1:Draft peek for \"{self.peekable_bough3}\" old ring: 0.0",
				case_loggy.output
			)








	def test_draft_picky_2(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.DRAFT_PEEK_HANDLER
				init_name	= "picky_2"
				init_level	= 10

			@DraftPeek(renew=False, cache=False, picky=True)
			class grow(Transmutable):
				def __call__(*args):	pass


		self.test_case = Sakura()
		self.fmake(self.peekable_bough1)
		bough1_peek = os.path.getmtime(self.peekable_bough1)
		self.fmake(self.peekable_bough2)
		bough2_peek = os.path.getmtime(self.peekable_bough2)
		self.fmake(self.peekable_bough3)
		bough3_peek = os.path.getmtime(self.peekable_bough3)

		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		first_peek = os.path.getmtime(self.peekable_sprout1)


		with self.assertLogs("picky_2", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:picky_2:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:picky_2:Draft peek for \"{self.peekable_bough1}\" old ring: {bough1_peek}",
				case_loggy.output
			)


		with self.assertLogs("picky_2", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:picky_2:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:picky_2:Draft peek for \"{self.peekable_bough2}\" old ring: {bough2_peek}",
				case_loggy.output
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		second_peek = os.path.getmtime(self.peekable_sprout1)
		self.assertNotEqual(first_peek, second_peek)


		with self.assertLogs("picky_2", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:picky_2:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{second_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:picky_2:Draft peek for \"{self.peekable_bough3}\" old ring: {bough3_peek}",
				case_loggy.output
			)








	def test_draft_picky_renew_1(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.DRAFT_PEEK_HANDLER
				init_name	= "picky_renew_1"
				init_level	= 10

			@DraftPeek(cache=False, picky=True)
			class grow(Transmutable):
				def __call__(*args):	pass


		self.test_case = Sakura()
		self.fmake(self.peekable_sprout1)
		first_peek = os.path.getmtime(self.peekable_sprout1)
		if	os.path.isfile(self.peekable_bough1): os.remove(self.peekable_bough1)
		if	os.path.isfile(self.peekable_bough2): os.remove(self.peekable_bough2)
		if	os.path.isfile(self.peekable_bough3): os.remove(self.peekable_bough3)


		with self.assertLogs("picky_renew_1", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:picky_renew_1:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:picky_renew_1:Flourish stopped cause draft peek only for sprigs renew",
				case_loggy.output
			)


		with self.assertLogs("picky_renew_1", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:picky_renew_1:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:picky_renew_1:Flourish stopped cause draft peek only for sprigs renew",
				case_loggy.output
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		second_peek = os.path.getmtime(self.peekable_sprout1)
		self.assertNotEqual(first_peek, second_peek)


		with self.assertLogs("picky_renew_1", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:picky_renew_1:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{second_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:picky_renew_1:Flourish stopped cause draft peek only for sprigs renew",
				case_loggy.output
			)








	def test_draft_picky_renew_2(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.DRAFT_PEEK_HANDLER
				init_name	= "picky_renew_2"
				init_level	= 10

			@DraftPeek(cache=False, picky=True)
			class grow(Transmutable):
				def __call__(*args):	pass


		self.test_case = Sakura()
		self.fmake(self.peekable_bough1)
		bough1_peek = os.path.getmtime(self.peekable_bough1)
		self.fmake(self.peekable_bough2)
		bough2_peek = os.path.getmtime(self.peekable_bough2)
		self.fmake(self.peekable_bough3)
		bough3_peek = os.path.getmtime(self.peekable_bough3)

		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		first_peek = os.path.getmtime(self.peekable_sprout1)


		with self.assertLogs("picky_renew_2", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:picky_renew_2:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:picky_renew_2:Draft peek for \"{self.peekable_bough1}\" old ring: {bough1_peek}",
				case_loggy.output
			)


		with self.assertLogs("picky_renew_2", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:picky_renew_2:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:picky_renew_2:Draft peek for \"{self.peekable_bough2}\" old ring: {bough2_peek}",
				case_loggy.output
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		second_peek = os.path.getmtime(self.peekable_sprout1)
		self.assertNotEqual(first_peek, second_peek)


		with self.assertLogs("picky_renew_2", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:picky_renew_2:New draft peek \"{self.peekable_sprout1}\" new ring: "
				f"{second_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:picky_renew_2:Draft peek for \"{self.peekable_bough3}\" old ring: {bough3_peek}",
				case_loggy.output
			)








	def test_draft_comparators(self):

		self.fmake(self.peekable_bough1)
		bough_peek = int(os.path.getmtime(self.peekable_bough1))

		sleep(1.1)
		self.fmake(self.peekable_sprout1)

		comparisons = {

			"Flourish stopped by draft peek comparator":							lambda A,B : A <B,
			f"Draft peek for \"{self.peekable_bough1}\" old ring: {bough_peek}":	lambda A,B : B <A,
			f"Draft peek for \"{self.peekable_bough1}\" old ring: {bough_peek}":	lambda A,B : A <= B,
			f"Draft peek for \"{self.peekable_bough1}\" old ring: {bough_peek}":	lambda A,B : B <= A,
			"Flourish stopped by draft peek comparator":							lambda A,B : A == B,
			f"Draft peek for \"{self.peekable_bough1}\" old ring: {bough_peek}":	lambda A,B : A != B,
		}

		for i,( message,comparator ) in enumerate(comparisons.items()):
			with self.assertLogs(f"comparator-{i}", 10) as case_loggy:

				class Sakura(Transmutable):
					class loggy(LibraryContrib):

						handler		= self.DRAFT_PEEK_HANDLER
						init_name	= f"comparator-{i}"
						init_level	= 10

					@DraftPeek(renew=False, comparator=comparator)
					class grow(Transmutable):
						def __call__(*args):	pass

				self.test_case = Sakura()
				self.test_case.grow(

					self.test_case,
					self.INIT_SET_SPROUT,
					Path(self.INIT_SET_SPROUT),
					Path(self.peekable_sprout1),
					Path(self.INIT_SET_BOUGH_1)
				)
				self.assertIn(f"DEBUG:comparator-{i}:{message}", case_loggy.output)
			self.test_case.loggy.close()








	def test_draft_picky_comparators(self):

		self.fmake(self.peekable_bough1)
		bough_peek = os.path.getmtime(self.peekable_bough1)

		sleep(1.1)
		self.fmake(self.peekable_sprout1)

		comparisons = {

			"Flourish stopped by draft peek comparator":							lambda A,B : A <B,
			f"Draft peek for \"{self.peekable_bough1}\" old ring: {bough_peek}":	lambda A,B : B <A,
			f"Draft peek for \"{self.peekable_bough1}\" old ring: {bough_peek}":	lambda A,B : A <= B,
			f"Draft peek for \"{self.peekable_bough1}\" old ring: {bough_peek}":	lambda A,B : B <= A,
			"Flourish stopped by draft peek comparator":							lambda A,B : A == B,
			f"Draft peek for \"{self.peekable_bough1}\" old ring: {bough_peek}":	lambda A,B : A != B,
		}

		for i,( message,comparator ) in enumerate(comparisons.items()):
			with self.assertLogs(f"picky_comparator-{i}", 10) as case_loggy:

				class Sakura(Transmutable):
					bough = self.INIT_SET_BOUGH_1
					class loggy(LibraryContrib):

						handler		= self.DRAFT_PEEK_HANDLER
						init_name	= f"picky_comparator-{i}"
						init_level	= 10

					@DraftPeek(renew=False, picky=True, comparator=comparator)
					class grow(Transmutable):
						def __call__(*args):	pass

				self.test_case = Sakura()
				self.test_case.grow(

					self.test_case,
					self.INIT_SET_SPROUT,
					Path(self.INIT_SET_SPROUT),
					Path(self.peekable_sprout1),
					Path(self.INIT_SET_BOUGH_1)
				)
				self.assertIn(f"DEBUG:picky_comparator-{i}:{message}", case_loggy.output)
			self.test_case.loggy.close()








	def test_draft_time_comparators(self):

		self.fmake(self.peekable_sprout1)
		self.fmake(self.peekable_bough1)

		sprout_peek = int(os.path.getmtime(self.peekable_sprout1))
		bough_peek = int(os.path.getmtime(self.peekable_bough1))
		self.assertEqual(sprout_peek, bough_peek)

		comparisons = {

			f"Draft peek for \"{self.peekable_bough1}\" old ring: {bough_peek}":	(
				lambda A,B : TimeTurner(minutes=-1).epoch <A
			),
		}

		for i,( message,comparator ) in enumerate(comparisons.items()):
			with self.assertLogs(f"time_comparator-{i}", 10) as case_loggy:

				class Sakura(Transmutable):
					class loggy(LibraryContrib):

						handler		= self.DRAFT_PEEK_HANDLER
						init_name	= f"time_comparator-{i}"
						init_level	= 10

					@DraftPeek(renew=False, comparator=comparator)
					class grow(Transmutable):
						def __call__(*args):	pass

				self.test_case = Sakura()
				self.test_case.grow(

					self.test_case,
					self.INIT_SET_SPROUT,
					Path(self.INIT_SET_SPROUT),
					Path(self.peekable_sprout1),
					Path(self.INIT_SET_BOUGH_1)
				)
				self.assertIn(f"DEBUG:time_comparator-{i}:{message}", case_loggy.output)
			self.test_case.loggy.close()








	def test_draft_raising_comparators(self):

		comparators = (

			lambda		: True,
			lambda A	: bool(A),
			lambda A	: A **2,
			lambda A,B	: A +B,
			lambda A,B	: "B <A",
			lambda A,B,C: A +B +C,
			lambda A,B,C: bool(A +B +C),
		)

		for comparator in comparators:

			try:

				class Sakura(Transmutable):

					@DraftPeek(comparator=comparator)
					class raising_grow(Transmutable):
						def __call__(*args):	pass


			except	Exception as E:

				self.assertIsInstance(E, TypeError)
				self.assertEqual(
					str(E), "Draft peek comparator must be Callable[[ int|float, int|float ], bool]"
				)








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	os
import	unittest
from	typing								import Any
from 	pygwarts.magical.philosophers_stone	import Transmutable
from	pygwarts.magical.spells				import flagrate
from	pygwarts.hedwig.mail.letter			import ObjectField
from	pygwarts.hedwig.mail.letter			import StringField
from	pygwarts.hedwig.mail.letter			import TextField
from	pygwarts.hedwig.mail.letter			import AddressField
from	pygwarts.hedwig.mail.letter.fields	import SenderField
from	pygwarts.hedwig.mail.letter.fields	import RecipientField
from	pygwarts.hedwig.mail.letter.fields	import ccField
from	pygwarts.hedwig.mail.letter.fields	import bccField
from	pygwarts.hedwig.mail.letter.fields	import SubjectField
from	pygwarts.hedwig.mail.letter.fields	import BodyField
from	pygwarts.hedwig.mail.utils			import EMAIL_REGEX_PATTERN
from	pygwarts.hedwig.mail.utils			import EmailValidator
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.irma.access.utils			import TextWrapper
from	pygwarts.tests.hedwig				import HedwigTestCase








class MailUtilsCases(HedwigTestCase):

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.HEDWIG_MAIL_UTILS_HANDLER):

				cls.email_validator.loggy.close()
				os.remove(cls.HEDWIG_MAIL_UTILS_HANDLER)

	@classmethod
	def setUpClass(cls):

		class CurrentValidator(EmailValidator):
			class loggy(LibraryContrib):

				handler		= cls.HEDWIG_MAIL_UTILS_HANDLER
				init_name	= "EmailValidator"
				init_level	= 10
		cls.email_validator = CurrentValidator()
		cls.make_loggy_file(cls, cls.HEDWIG_MAIL_UTILS_HANDLER)


	def test_EMAIL_REGEX_PATTERN_valid(self):

		for address in self.VALID_EMAILS : self.assertTrue(EMAIL_REGEX_PATTERN.fullmatch(address))

	def test_EMAIL_REGEX_PATTERN_invalid(self):

		for address in self.INVALID_EMAILS : self.assertIsNone(EMAIL_REGEX_PATTERN.fullmatch(address))

	def test_EmailValidator_valid(self):

		for address in self.VALID_EMAILS:

			with self.subTest(email=address):
				with self.assertLogs("EmailValidator", 10) as case_loggy:

					self.assertTrue(self.email_validator(address))
				self.assertIn(
					f"DEBUG:EmailValidator:Email address \"{address}\" validated", case_loggy.output
				)


	def test_EmailValidator_invalid(self):

		for address in (

			*self.INVALID_EMAILS,
			1, 1., 1E1, True, False, None, ..., print, type,
			[ "email@example.com" ],
			( "email@example.com", ),
			{ "email@example.com" },
			{ "email": "email@example.com" },
		):
			with self.subTest(email=address):
				with self.assertLogs("EmailValidator", 20) as case_loggy:

					self.assertFalse(self.email_validator(address))
				self.assertIn(f"INFO:EmailValidator:Invalid email address \"{address}\"", case_loggy.output)








	def test_TextWrapper_Transmutable(self):

		@TextWrapper("this is ", ", baby!")
		class TextGenerator(Transmutable):

			def __call__(self): return "generated text"
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_UTILS_HANDLER
				init_name	= "TextWrapper_Transmutable"
				init_level	= 10

		with self.assertLogs("TextWrapper_Transmutable", 10) as case_loggy:

			self.test_case = TextGenerator()
			current = self.test_case()

		self.assertEqual(len(case_loggy.output),1)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertIn("DEBUG:TextWrapper_Transmutable:Wrapped text now 29 symbols", case_loggy.output)
		self.assertEqual(current, "this is generated text, baby!")




	def test_TextWrapper_Transmutable_args(self):

		@TextWrapper("this is ", ", baby!")
		class TextGenerator(Transmutable):

			def __call__(self, what :Any): return f"generated {what}"
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_UTILS_HANDLER
				init_name	= "TextWrapper_Transmutable_args"
				init_level	= 10

		self.test_case = TextGenerator()
		for arg in (

			"text",( "text", ),[ "text" ],{ "text" },{ "what": "text" },
			None, print, ..., 1, .1, True, False, Transmutable, self.test_case
		):
			with self.subTest(arg=arg):
				with self.assertLogs("TextWrapper_Transmutable_args", 10) as case_loggy:

					current = self.test_case(arg)

				L = len(current)
				self.assertEqual(len(case_loggy.output),1)
				self.no_loggy_levels(case_loggy.output, 30,40,50)
				self.assertIn(

					f"DEBUG:TextWrapper_Transmutable_args:Wrapped text now {L} symbol{flagrate(L)}",
					case_loggy.output
				)
				self.assertEqual(current, f"this is generated {arg}, baby!")




	def test_TextWrapper_Transmutable_invalid_args(self):

		@TextWrapper("this is ", ", baby!")
		class TextGenerator(Transmutable):

			def __call__(self, what :Any): return what
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_UTILS_HANDLER
				init_name	= "Transmutable_invalid_args"
				init_level	= 10

		self.test_case = TextGenerator()
		for arg in (

			( "text", ),[ "text" ],{ "text" },{ "what": "text" },
			None, print, ..., 1, .1, True, False, Transmutable, self.test_case
		):
			with self.subTest(arg=arg):
				with self.assertLogs("Transmutable_invalid_args", 10) as case_loggy:

					current = self.test_case(arg)

				self.assertEqual(len(case_loggy.output),1)
				self.no_loggy_levels(case_loggy.output, 30,40,50)
				self.assertIn(

					f"DEBUG:Transmutable_invalid_args:Text value to wrap not found",
					case_loggy.output
				)
				self.assertIsNone(current)








	def test_TextWrapper_ObjectField(self):

		@TextWrapper("this is ", ", baby!")
		class TextGenerator(ObjectField):

			field_value = "generated text"
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_UTILS_HANDLER
				init_name	= "TextWrapper_ObjectField"
				init_level	= 10

		with self.assertLogs("TextWrapper_ObjectField", 10) as case_loggy:

			self.test_case = TextGenerator()
			current = self.test_case()

		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertIn(f"DEBUG:TextWrapper_ObjectField:Fetched value of {type('')}", case_loggy.output)
		self.assertIn("DEBUG:TextWrapper_ObjectField:Wrapped text now 29 symbols", case_loggy.output)
		self.assertEqual(current, "this is generated text, baby!")




	def test_TextWrapper_ObjectField_fail(self):

		@TextWrapper("this is ", ", baby!")
		class TextGenerator(ObjectField):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_UTILS_HANDLER
				init_name	= "TextWrapper_ObjectField_fail"
				init_level	= 10

		with self.assertLogs("TextWrapper_ObjectField_fail", 10) as case_loggy:

			self.test_case = TextGenerator()
			current = self.test_case()

		self.assertEqual(len(case_loggy.output),2)
		self.no_loggy_levels(case_loggy.output, 50)
		self.assertIn(
			f"ERROR:TextWrapper_ObjectField_fail:Field {self.test_case} must have value", case_loggy.output
		)
		self.assertIn("DEBUG:TextWrapper_ObjectField_fail:Text value to wrap not found", case_loggy.output)
		self.assertIsNone(current)








	def test_TextWrapper_StringField(self):

		@TextWrapper("this is ", ", baby!")
		class TextGenerator(StringField):

			field_value = "generated text"
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_UTILS_HANDLER
				init_name	= "TextWrapper_StringField"
				init_level	= 10

		with self.assertLogs("TextWrapper_StringField", 10) as case_loggy:

			self.test_case = TextGenerator()
			current = self.test_case()

		self.assertEqual(len(case_loggy.output),3)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertIn(f"DEBUG:TextWrapper_StringField:Fetched value of {type('')}", case_loggy.output)
		self.assertIn(f"DEBUG:TextWrapper_StringField:Fetched 14 symbols", case_loggy.output)
		self.assertIn("DEBUG:TextWrapper_StringField:Wrapped text now 29 symbols", case_loggy.output)
		self.assertEqual(current, "this is generated text, baby!")




	def test_TextWrapper_StringField_fail(self):

		@TextWrapper("this is ", ", baby!")
		class TextGenerator(StringField):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_UTILS_HANDLER
				init_name	= "TextWrapper_StringField_fail"
				init_level	= 10

		with self.assertLogs("TextWrapper_StringField_fail", 10) as case_loggy:

			self.test_case = TextGenerator()
			current = self.test_case()

		self.assertEqual(len(case_loggy.output),3)
		self.no_loggy_levels(case_loggy.output, 50)
		self.assertIn(
			f"ERROR:TextWrapper_StringField_fail:Field {self.test_case} must have value", case_loggy.output
		)
		self.assertIn(f"DEBUG:TextWrapper_StringField_fail:No value fetched for field", case_loggy.output)
		self.assertIn("DEBUG:TextWrapper_StringField_fail:Text value to wrap not found", case_loggy.output)
		self.assertIsNone(current)








	def test_TextWrapper_TextField(self):

		@TextWrapper("this is ", ", baby!")
		class TextGenerator(TextField):

			field_value = "generated text"
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_UTILS_HANDLER
				init_name	= "TextWrapper_TextField"
				init_level	= 10

		with self.assertLogs("TextWrapper_TextField", 10) as case_loggy:

			self.test_case = TextGenerator()
			current = self.test_case()

		self.assertEqual(len(case_loggy.output),5)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertIn(f"DEBUG:TextWrapper_TextField:Fetched value of {type('')}", case_loggy.output)
		self.assertIn(f"DEBUG:TextWrapper_TextField:Fetched 14 symbols", case_loggy.output)
		self.assertIn(f"DEBUG:TextWrapper_TextField:Proper modifiers not found", case_loggy.output)
		self.assertIn(f"DEBUG:TextWrapper_TextField:Obtained 14 symbols", case_loggy.output)
		self.assertIn("DEBUG:TextWrapper_TextField:Wrapped text now 29 symbols", case_loggy.output)
		self.assertEqual(current, "this is generated text, baby!")




	def test_TextWrapper_TextField_modifiers(self):

		@TextWrapper("this is ", ", baby!")
		class TextGenerator(TextField):

			field_value	= "generated {what}"
			modifiers	= { "what": "text" }
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_UTILS_HANDLER
				init_name	= "TextWrapper_TextField_modifiers"
				init_level	= 10

		with self.assertLogs("TextWrapper_TextField_modifiers", 10) as case_loggy:

			self.test_case = TextGenerator()
			current = self.test_case()

		self.assertEqual(len(case_loggy.output),4)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertIn(f"DEBUG:TextWrapper_TextField_modifiers:Fetched value of {type('')}", case_loggy.output)
		self.assertIn(f"DEBUG:TextWrapper_TextField_modifiers:Fetched 16 symbols", case_loggy.output)
		self.assertIn(f"DEBUG:TextWrapper_TextField_modifiers:Obtained 14 symbols", case_loggy.output)
		self.assertIn("DEBUG:TextWrapper_TextField_modifiers:Wrapped text now 29 symbols", case_loggy.output)
		self.assertEqual(current, "this is generated text, baby!")




	def test_TextWrapper_TextField_fail(self):

		@TextWrapper(header="this is ", footer=", baby!")
		class TextGenerator(TextField):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_UTILS_HANDLER
				init_name	= "TextWrapper_TextField_fail"
				init_level	= 10

		with self.assertLogs("TextWrapper_TextField_fail", 10) as case_loggy:

			self.test_case = TextGenerator()
			current = self.test_case()

		self.assertEqual(len(case_loggy.output),3)
		self.no_loggy_levels(case_loggy.output, 50)
		self.assertIn(
			f"ERROR:TextWrapper_TextField_fail:Field {self.test_case} must have value", case_loggy.output
		)
		self.assertIn(f"DEBUG:TextWrapper_TextField_fail:No value fetched for field", case_loggy.output)
		self.assertIn("DEBUG:TextWrapper_TextField_fail:Text value to wrap not found", case_loggy.output)
		self.assertIsNone(current)




	def test_TextWrapper_TextField_modifiers_fail(self):

		@TextWrapper(footer=", baby!", header="this is ")
		class TextGenerator(TextField):

			field_value	= "generated {what}"
			modifiers	= { "where": "text" }
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_UTILS_HANDLER
				init_name	= "TextWrapper_TextField_modifiers_fail"
				init_level	= 10

		with self.assertLogs("TextWrapper_TextField_modifiers_fail", 10) as case_loggy:

			self.test_case = TextGenerator()
			current = self.test_case()

		self.assertEqual(len(case_loggy.output),4)
		self.no_loggy_levels(case_loggy.output, 50)
		self.assertIn(

			f"DEBUG:TextWrapper_TextField_modifiers_fail:Fetched value of {type('')}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:TextWrapper_TextField_modifiers_fail:Fetched 16 symbols",
			case_loggy.output
		)
		self.assertIn(

			f"ERROR:TextWrapper_TextField_modifiers_fail:{self.test_case} field failed due to "
			"KeyError: 'what'",
			case_loggy.output
		)
		self.assertIn(

			"DEBUG:TextWrapper_TextField_modifiers_fail:Text value to wrap not found",
			case_loggy.output
		)
		self.assertIsNone(current)








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	re
from	pygwarts.filch.nettherin	import VALID_IP4








VALID_MAC = str(

	r"(([A-Fa-f0-9]{2}-){5}([A-Fa-f0-9]{2}))|"
	r"(([A-Fa-f0-9]{2}:){5}([A-Fa-f0-9]{2}))|"
	r"(([A-Fa-f0-9]{4}\.){2}[A-Fa-f0-9]{4})"
)
G_O1_VALID_MAC = str(

	r"(?P<octet1>[A-Fa-f0-9]{2})[:-]"
	r"(?P<octet2>[A-Fa-f0-9]{2})[:-]"
	r"(?P<octet3>[A-Fa-f0-9]{2})[:-]"
	r"(?P<octet4>[A-Fa-f0-9]{2})[:-]"
	r"(?P<octet5>[A-Fa-f0-9]{2})[:-]"
	r"(?P<octet6>[A-Fa-f0-9]{2})"
)
G_O2_VALID_MAC	= r"(?P<octets1>[A-Fa-f0-9]{4})\.(?P<octets2>[A-Fa-f0-9]{4})\.(?P<octets3>[A-Fa-f0-9]{4})"
P_VALID_MAC		= re.compile(VALID_MAC)
GP_O1_VALID_MAC	= re.compile(G_O1_VALID_MAC)
GP_O2_VALID_MAC	= re.compile(G_O2_VALID_MAC)
P_ARP_REQ		= re.compile(f"[Ww]ho has {VALID_IP4} says {VALID_IP4}")
GP_ARP_REQ		= re.compile(fr"[Ww]ho has (?P<dst>{VALID_IP4}) says (?P<src>{VALID_IP4})")
GP_ARP_REQ_LOG	= re.compile(fr"{GP_ARP_REQ.pattern} \((?P<mac>{VALID_MAC})\)")
GP_ARP_RES_LOG	= re.compile(fr"(?P<mac>{VALID_MAC}) answered")








def is_valid_MAC(addr :str) -> bool | None :

	"""
		Helper function that verifies is "addr" string argument a valid MAC address record. Verification
		goes the classic way by splitting to parts and validating it is one or two octets, according to IEEE
		EUI-48. Returns True if "addr" consists of exactly six valid octets, delimited according to IEEE
		EUI-48. Returns False if at least one part is invalid, according to IEEE EUI-48, or if it is not
		one or two valid octets part. Returns None for the rest.
	"""

	if	isinstance(addr, str):

		try:

			if	len(S := addr.split("-")) == 6 or len(S := addr.split(":")) == 6:
				return	all( 0 <= int(octet,16) <256 for octet in S )

			elif(len(S := addr.split(".")) == 3):
				return	all( 0 <= int(octets,16) <65536 for octets in S )
		except:	return




def EUI48_format(addr :str, d :str ="-") -> str | None | Exception :

	"""
		Utility function that accepts "addr" and "d" strings and does two things:
			1. verifies "addr" is a valid MAC address record string;
			2. converts "addr" to a specified by "d" delimiter char EUI-48 MAC format.
		EUI-48 suggests three types of MAC records, so this is possible formats to be returned:
			XX-XX-XX-XX-XX-XX
			XX:XX:XX:XX:XX:XX
			XXXX.XXXX.XXXX
		Handles no Exception that might be raised and raises AssertionError if "d" is invalid
		delimiter. Returns None if "addr" failed validation.
	"""


	assert	str(d) in "-:.", f"EUI-48 delimiter must be \":\", \"-\" or \".\", not \"{d}\""


	if	(match := GP_O1_VALID_MAC.fullmatch(str(addr))):
		o1,o2,o3,o4,o5,o6 = match.group("octet1","octet2","octet3","octet4","octet5","octet6")

		if	all(filter(bool,( o1,o2,o3,o4,o5,o6 ))):
			return	(

				f"{o1}{d}{o2}{d}{o3}{d}{o4}{d}{o5}{d}{o6}".lower()
				if	d in "-:" else
				f"{o1}{o2}.{o3}{o4}.{o5}{o6}".lower()
			)


	elif(match := GP_O2_VALID_MAC.fullmatch(str(addr))):
		os1,os2,os3 = match.group("octets1","octets2","octets3")

		if	all(filter(bool,( os1,os2,os3 ))):
			return	(

				f"{os1[:2]}{d}{os1[2:]}{d}{os2[:2]}{d}{os2[2:]}{d}{os3[:2]}{d}{os3[2:]}".lower()
				if	d in "-:" else
				f"{os1}.{os2}.{os3}".lower()
			)










from pygwarts.irma.access.volume	import LibraryVolume
from pygwarts.irma.access.handlers	import AccessHandler








class AccessCounter(AccessHandler):

	"""
		Implementation of handler, that upon invocation does only two things:
			1. AccessHandlerCounter job - increments "access_handler_counter" field;
			2. AccessHandlerRegisterCounter job - increments it's "counter" integer mapping in
			LibraryVolume object "volume".
		Always returns True.
	"""

	def __call__(self, line :str, volume :LibraryVolume) -> True :

		self.access_handler_counter = getattr(self, "access_handler_counter", 0) +1
		self.loggy.debug(f"Handler counter incremented to {self.access_handler_counter}")

		if	self.registered(volume):

			volume[self].setdefault("counter",0)
			volume[self]["counter"] += 1
			self.loggy.debug(f"{volume} counter incremented to {volume[self]['counter']}")


		return True










from pathlib		import Path
from typing			import List
from typing			import Tuple
from typing			import Generator
from collections	import deque
from itertools		import filterfalse








def fstree(root :str | Path) -> Generator[Tuple[Path, List[Path], List[Path]] | List, None,None] :

	"""
		hagrid core generator, that represents iterative analogue for standard library os.walk.
		Not restricted by recursion limit and uses breadth-first algorithm, so must safe some memory. Also
		"deque" as a maintainable data structure must safe some time.
		Accepts the only argument "root" which must be a string that represents the absolute path for a
		folder to start iterative "walk" with, or a corresponding Path object. In both cases a new Path
		object will be created and used. In case of invalid "root" generator yields empty list and stops.
		When valid root provided it is putted in "deque" and the "walk" will start. The breadth of the "walk"
		is achieved by popping current "deque" and pushing new Path object, that represent subfolders, to the
		left. Every popped item must be a Path object and represent "root" subfolder. Every popped item, as a
		folder, will be iterated to make two lists of it's items - list of folders and list of files. Those
		two lists will follow current subfolder as a three objects tuple that will be yielded. In any case,
		when popped item is not a Path object that represents a subfolder, or it cannot produce somehow a
		list of Path objects by invocation of it's (perhaps absent in this case) "iterdir" Path object method,
		generator will yield an empty list and continue. Thus this generator always yields.
	"""

	try:	walk	= deque([ Path(root) ])
	except:	yield	from list()
	else:

		while	walk:

			for _ in range(len(walk)):
				branch	= walk.pop()

				try:

					sprigs	= list(branch.iterdir())
					twigs	= list(filterfalse(Path.is_symlink, filter(Path.is_dir, sprigs)))
					leafs	= list(filterfalse(Path.is_symlink, filter(Path.is_file,sprigs)))
					yield	branch, twigs, leafs
					for		twig in twigs : walk.appendleft(twig)
				except:		yield list()







